# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re
import os
import mock
import signal
from twisted.internet import defer, reactor, task
from twisted.trial import unittest
from twisted.python import log
from buildbot import master, monkeypatches, config
from buildbot.util import subscription
from buildbot.db import connector
from buildbot.test.util import dirs, compat, misc
from buildbot.test.fake import fakedb
from buildbot.util import epoch2datetime
from buildbot.changes import changes
from buildbot.process.users import users

class Subscriptions(dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        basedir = os.path.abspath('basedir')
        d = self.setUpDirs(basedir)
        def set_master(_):
            self.master = master.BuildMaster(basedir)
            self.master.config.db['db_poll_interval'] = None
        d.addCallback(set_master)
        return d

    def tearDown(self):
        return self.tearDownDirs()

    def test_change_subscription(self):
        changeid = 918
        chdict = {
            'changeid': 14,
            'author': u'warner',
            'branch': u'warnerdb',
            'category': u'devel',
            'comments': u'fix whitespace',
            'files': [u'master/buildbot/__init__.py'],
            'is_dir': 0,
            'project': u'Buildbot',
            'properties': {},
            'repository': u'git://warner',
            'revision': u'0e92a098b',
            'revlink': u'http://warner/0e92a098b',
            'when_timestamp': epoch2datetime(266738404),
        }
        newchange = mock.Mock(name='newchange')

        # patch out everything we're about to call
        self.master.db = mock.Mock()
        self.master.db.changes.addChange.return_value = \
            defer.succeed(changeid)
        self.master.db.changes.getChange.return_value = \
            defer.succeed(chdict)
        self.patch(changes.Change, 'fromChdict',
                classmethod(lambda cls, master, chdict :
                                defer.succeed(newchange)))

        cb = mock.Mock()
        sub = self.master.subscribeToChanges(cb)
        self.assertIsInstance(sub, subscription.Subscription)

        d = self.master.addChange()
        def check(change):
            # master called the right thing in the db component, including with
            # appropriate default values
            self.master.db.changes.addChange.assert_called_with(author=None,
                    files=None, comments=None, is_dir=0,
                    revision=None, when_timestamp=None, branch=None,
                    category=None, revlink='', properties={}, repository='',
                    project='', uid=None)

            self.master.db.changes.getChange.assert_called_with(changeid)
            # addChange returned the right value
            self.failUnless(change is newchange) # fromChdict's return value
            # and the notification sub was called correctly
            cb.assert_called_with(newchange)
        d.addCallback(check)
        return d

    def do_test_addChange_args(self, args=(), kwargs={}, exp_db_kwargs={}):
        # add default arguments
        default_db_kwargs = dict(files=None, comments=None, author=None,
                is_dir=0, revision=None, when_timestamp=None,
                branch=None, category=None, revlink='', properties={},
                repository='', project='', uid=None)
        k = default_db_kwargs
        k.update(exp_db_kwargs)
        exp_db_kwargs = k

        self.master.db = mock.Mock()
        got = []
        def db_addChange(*args, **kwargs):
            got[:] = args, kwargs
            # use an exception as a quick way to bail out of the remainder
            # of the addChange method
            return defer.fail(RuntimeError)
        self.master.db.changes.addChange = db_addChange

        d = self.master.addChange(*args, **kwargs)
        d.addCallback(lambda _ : self.fail("should not succeed"))
        def check(f):
            self.assertEqual(got, [(), exp_db_kwargs])
        d.addErrback(check)
        return d

    def test_addChange_args_author(self):
        # who should come through as author
        return self.do_test_addChange_args(
                kwargs=dict(who='me'),
                exp_db_kwargs=dict(author='me'))

    def test_addChange_args_isdir(self):
        # isdir should come through as is_dir
        return self.do_test_addChange_args(
                kwargs=dict(isdir=1),
                exp_db_kwargs=dict(is_dir=1))

    def test_addChange_args_when(self):
        # when should come through as when_timestamp, as a datetime
        return self.do_test_addChange_args(
                kwargs=dict(when=892293875),
                exp_db_kwargs=dict(when_timestamp=epoch2datetime(892293875)))

    def test_addChange_args_properties(self):
        # properties should be qualified with a source
        return self.do_test_addChange_args(
                kwargs=dict(properties={ 'a' : 'b' }),
                exp_db_kwargs=dict(properties={ 'a' : ('b', 'Change') }))

    def test_addChange_args_properties_tuple(self):
        # properties should be qualified with a source, even if they
        # already look like they have a source
        return self.do_test_addChange_args(
                kwargs=dict(properties={ 'a' : ('b', 'Change') }),
                exp_db_kwargs=dict(properties={
                    'a' : (('b', 'Change'), 'Change') }))

    def test_addChange_args_positional(self):
        # master.addChange can take author, files, comments as positional
        # arguments
        return self.do_test_addChange_args(
                args=('me', ['a'], 'com'),
                exp_db_kwargs=dict(author='me', files=['a'], comments='com'))

    def do_test_createUserObjects_args(self, args=(), kwargs={}, exp_args=()):
        got = []
        def fake_createUserObject(*args, **kwargs):
            got[:] = args, kwargs
            # use an exception as a quick way to bail out of the remainder
            # of the createUserObject method
            return defer.fail(RuntimeError)

        self.patch(users, 'createUserObject', fake_createUserObject)

        d = self.master.addChange(*args, **kwargs)
        d.addCallback(lambda _ : self.fail("should not succeed"))
        def check(f):
            self.assertEqual(got, [exp_args, {}])
        d.addErrback(check)
        return d

    def test_addChange_createUserObject_args(self):
        # who should come through as author
        return self.do_test_createUserObjects_args(
                kwargs=dict(who='me', src='git'),
                exp_args=(self.master, 'me', 'git'))

    def test_buildset_subscription(self):
        self.master.db = mock.Mock()
        self.master.db.buildsets.addBuildset.return_value = \
            defer.succeed((938593, dict(a=19,b=20)))

        cb = mock.Mock()
        sub = self.master.subscribeToBuildsets(cb)
        self.assertIsInstance(sub, subscription.Subscription)

        d = self.master.addBuildset(ssid=999)
        def check((bsid,brids)):
            # master called the right thing in the db component
            self.master.db.buildsets.addBuildset.assert_called_with(ssid=999)
            # addBuildset returned the right value
            self.assertEqual((bsid,brids), (938593, dict(a=19,b=20)))
            # and the notification sub was called correctly
            cb.assert_called_with(bsid=938593, ssid=999)
        d.addCallback(check)
        return d

    def test_buildset_completion_subscription(self):
        self.master.db = mock.Mock()

        cb = mock.Mock()
        sub = self.master.subscribeToBuildsetCompletions(cb)
        self.assertIsInstance(sub, subscription.Subscription)

        self.master._buildsetComplete(938593, 999)
        # assert the notification sub was called correctly
        cb.assert_called_with(938593, 999)

class StartupAndReconfig(dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        self.basedir = os.path.abspath('basedir')
        d = self.setUpDirs(self.basedir)
        @d.addCallback
        def make_master(_):
            # don't create child services
            self.patch(master.BuildMaster, 'create_child_services',
                    lambda self : None)

            # patch out a few other annoying things the msater likes to do
            self.patch(monkeypatches, 'patch_all', lambda : None)
            self.patch(signal, 'signal', lambda sig, hdlr : None)
            self.patch(master, 'Status', lambda master : mock.Mock()) # XXX temporary
            self.patch(config.MasterConfig, 'loadConfig',
                    classmethod(lambda cls, b, f : cls()))

            self.master = master.BuildMaster(self.basedir)
            self.db = self.master.db = fakedb.FakeDBConnector(self)

        @d.addCallback
        def patch_log_msg(_):
            msg = mock.Mock(side_effect=log.msg)
            self.patch(log, 'msg', msg)

        return d

    def tearDown(self):
        return self.tearDownDirs()

    def make_reactor(self):
        r = mock.Mock()
        r.callWhenRunning = reactor.callWhenRunning
        return r

    def assertLogged(self, regexp):
        r = re.compile(regexp)
        for args, kwargs in log.msg.call_args_list:
            if args and r.search(args[0]):
                return
        self.fail("%r not matched in log output" % regexp)

    def patch_loadConfig_fail(self):
        @classmethod
        def loadConfig(cls, b, f):
            config.error('oh noes')
        self.patch(config.MasterConfig, 'loadConfig', loadConfig)


    # tests

    def test_startup_bad_config(self):
        reactor = self.make_reactor()
        self.patch_loadConfig_fail()

        d = self.master.startService(_reactor=reactor)
        d.addCallback(lambda _ : self.master.stopService())

        @d.addCallback
        def check(_):
            reactor.stop.assert_called()
            self.assertLogged("oh noes")
        return d

    def test_startup_db_not_ready(self):
        reactor = self.make_reactor()
        def db_setup():
            log.msg("GOT HERE")
            raise connector.DatabaseNotReadyError()
        self.db.setup = db_setup

        d = self.master.startService(_reactor=reactor)
        d.addCallback(lambda _ : self.master.stopService())

        @d.addCallback
        def check(_):
            reactor.stop.assert_called()
            self.assertLogged("GOT HERE")
        return d

    @compat.usesFlushLoggedErrors
    def test_startup_error(self):
        reactor = self.make_reactor()
        def db_setup():
            raise RuntimeError("oh noes")
        self.db.setup = db_setup

        d = self.master.startService(_reactor=reactor)
        d.addCallback(lambda _ : self.master.stopService())

        @d.addCallback
        def check(_):
            reactor.stop.assert_called()
            self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)
        return d

    def test_startup_ok(self):
        reactor = self.make_reactor()

        d = self.master.startService(_reactor=reactor)
        d.addCallback(lambda _ : self.master.stopService())

        @d.addCallback
        def check(_):
            self.failIf(reactor.stop.called)
            self.assertLogged("BuildMaster is running")
        return d

    def test_reconfig(self):
        reactor = self.make_reactor()
        self.master.reconfigService = mock.Mock(
                side_effect=lambda n : defer.succeed(None))

        d = self.master.startService(_reactor=reactor)
        d.addCallback(lambda _ : self.master.reconfig())
        d.addCallback(lambda _ : self.master.stopService())

        @d.addCallback
        def check(_):
            self.master.reconfigService.assert_called()
        return d

    @defer.deferredGenerator
    def test_reconfig_bad_config(self):
        reactor = self.make_reactor()
        self.master.reconfigService = mock.Mock(
                side_effect=lambda n : defer.succeed(None))

        wfd = defer.waitForDeferred(
            self.master.startService(_reactor=reactor))
        yield wfd
        wfd.getResult()


        # reset, since startService called reconfigService
        self.master.reconfigService.reset_mock()

        # reconfig, with a failure
        self.patch_loadConfig_fail()
        wfd = defer.waitForDeferred(
            self.master.reconfig())
        yield wfd
        wfd.getResult()

        self.master.stopService()

        self.assertLogged("reconfig aborted without")
        self.failIf(self.master.reconfigService.called)

    def test_reconfigService_db_url_changed(self):
        old = self.master.config = config.MasterConfig()
        old.db['db_url'] = 'aaaa'
        new = config.MasterConfig()
        new.db['db_url'] = 'bbbb'

        self.assertRaises(config.ConfigErrors, lambda :
            self.master.reconfigService(new))

    def test_reconfigService_start_polling(self):
        loopingcall = mock.Mock()
        self.patch(task, 'LoopingCall', lambda fn : loopingcall)

        self.master.config = config.MasterConfig()
        new = config.MasterConfig()
        new.db['db_poll_interval'] = 120

        d = self.master.reconfigService(new)
        @d.addCallback
        def check(_):
            loopingcall.start.assert_called_with(120, now=False)
        return d

    def test_reconfigService_stop_polling(self):
        db_loop = self.master.db_loop = mock.Mock()

        old = self.master.config = config.MasterConfig()
        old.db['db_poll_interval'] = 120
        new = config.MasterConfig()
        new.db['db_poll_interval'] = None

        d = self.master.reconfigService(new)
        @d.addCallback
        def check(_):
            db_loop.stop.assert_called()
            self.assertEqual(self.master.db_loop, None)
        return d


class Polling(dirs.DirsMixin, misc.PatcherMixin, unittest.TestCase):

    def setUp(self):
        self.gotten_changes = []
        self.gotten_buildset_additions = []
        self.gotten_buildset_completions = []
        self.gotten_buildrequest_additions = []


        basedir = os.path.abspath('basedir')

        # patch out os.uname so that we get a consistent hostname
        self.patch_os_uname(lambda : [ 0, 'testhost.localdomain' ])
        self.master_name = "testhost.localdomain:%s" % (basedir,)

        d = self.setUpDirs(basedir)
        def set_master(_):
            self.master = master.BuildMaster(basedir)

            self.db = self.master.db = fakedb.FakeDBConnector(self)

            self.master.config.db['db_poll_interval'] = 10

            # overridesubscription callbacks
            self.master._change_subs = sub = mock.Mock()
            sub.deliver = self.deliverChange
            self.master._new_buildset_subs = sub = mock.Mock()
            sub.deliver = self.deliverBuildsetAddition
            self.master._complete_buildset_subs = sub = mock.Mock()
            sub.deliver = self.deliverBuildsetCompletion
            self.master._new_buildrequest_subs = sub = mock.Mock()
            sub.deliver = self.deliverBuildRequestAddition

        d.addCallback(set_master)
        return d

    def tearDown(self):
        return self.tearDownDirs()

    def deliverChange(self, change):
        self.gotten_changes.append(change)

    def deliverBuildsetAddition(self, **kwargs):
        self.gotten_buildset_additions.append(kwargs)

    def deliverBuildsetCompletion(self, bsid, result):
        self.gotten_buildset_completions.append((bsid, result))

    def deliverBuildRequestAddition(self, notif):
        self.gotten_buildrequest_additions.append(notif)

    # tests

    def test_pollDatabaseChanges_empty(self):
        self.db.insertTestData([
            fakedb.Object(id=22, name=self.master_name,
                          class_name='buildbot.master.BuildMaster'),
        ])
        d = self.master.pollDatabaseChanges()
        def check(_):
            self.assertEqual(self.gotten_changes, [])
            self.assertEqual(self.gotten_buildset_additions, [])
            self.assertEqual(self.gotten_buildset_completions, [])
            self.db.state.assertState(22, last_processed_change=0)
        d.addCallback(check)
        return d

    def test_pollDatabaseChanges_catchup(self):
        # with no existing state, it should catch up to the most recent change,
        # but not process anything
        self.db.insertTestData([
            fakedb.Object(id=22, name=self.master_name,
                          class_name='buildbot.master.BuildMaster'),
            fakedb.Change(changeid=10),
            fakedb.Change(changeid=11),
        ])
        d = self.master.pollDatabaseChanges()
        def check(_):
            self.assertEqual(self.gotten_changes, [])
            self.assertEqual(self.gotten_buildset_additions, [])
            self.assertEqual(self.gotten_buildset_completions, [])
            self.db.state.assertState(22, last_processed_change=11)
        d.addCallback(check)
        return d

    def test_pollDatabaseChanges_multiple(self):
        self.db.insertTestData([
            fakedb.Object(id=53, name=self.master_name,
                          class_name='buildbot.master.BuildMaster'),
            fakedb.ObjectState(objectid=53, name='last_processed_change',
                               value_json='10'),
            fakedb.Change(changeid=10),
            fakedb.Change(changeid=11),
            fakedb.Change(changeid=12),
        ])
        d = self.master.pollDatabaseChanges()
        def check(_):
            self.assertEqual([ ch.number for ch in self.gotten_changes],
                             [ 11, 12 ]) # note 10 was already seen
            self.assertEqual(self.gotten_buildset_additions, [])
            self.assertEqual(self.gotten_buildset_completions, [])
            self.db.state.assertState(53, last_processed_change=12)
        d.addCallback(check)
        return d

    def test_pollDatabaseChanges_nothing_new(self):
        self.db.insertTestData([
            fakedb.Object(id=53, name='master',
                          class_name='buildbot.master.BuildMaster'),
            fakedb.ObjectState(objectid=53, name='last_processed_change',
                               value_json='10'),
            fakedb.Change(changeid=10),
        ])
        d = self.master.pollDatabaseChanges()
        def check(_):
            self.assertEqual(self.gotten_changes, [])
            self.assertEqual(self.gotten_buildset_additions, [])
            self.assertEqual(self.gotten_buildset_completions, [])
            self.db.state.assertState(53, last_processed_change=10)
        d.addCallback(check)
        return d

    def test_pollDatabaseBuildRequests_empty(self):
        d = self.master.pollDatabaseBuildRequests()
        def check(_):
            self.assertEqual(self.gotten_buildrequest_additions, [])
        d.addCallback(check)
        return d

    def test_pollDatabaseBuildRequests_new(self):
        self.db.insertTestData([
            fakedb.SourceStampSet(id=127),
            fakedb.SourceStamp(id=127, sourcestampsetid=127),
            fakedb.Buildset(id=99, sourcestampsetid=127),
            fakedb.BuildRequest(id=19, buildsetid=99, buildername='9teen'),
            fakedb.BuildRequest(id=20, buildsetid=99, buildername='twenty')
        ])
        d = self.master.pollDatabaseBuildRequests()
        def check(_):
            self.assertEqual(sorted(self.gotten_buildrequest_additions),
                    sorted([dict(bsid=99, brid=19, buildername='9teen'),
                            dict(bsid=99, brid=20, buildername='twenty')]))
        d.addCallback(check)
        return d

    def test_pollDatabaseBuildRequests_incremental(self):
        d = defer.succeed(None)
        def insert1(_):
            self.db.insertTestData([
            fakedb.SourceStampSet(id=127),
            fakedb.SourceStamp(id=127, sourcestampsetid=127),
            fakedb.Buildset(id=99, sourcestampsetid=127),
            fakedb.BuildRequest(id=11, buildsetid=9, buildername='eleventy'),
            ])
        d.addCallback(insert1)
        d.addCallback(lambda _ : self.master.pollDatabaseBuildRequests())
        def insert2_and_claim(_):
            self.gotten_buildrequest_additions.append('MARK')
            self.db.insertTestData([
                fakedb.BuildRequest(id=20, buildsetid=9,
                                        buildername='twenty'),
            ])
            self.db.buildrequests.fakeClaimBuildRequest(11)
        d.addCallback(insert2_and_claim)
        d.addCallback(lambda _ : self.master.pollDatabaseBuildRequests())
        def unclaim(_):
            self.gotten_buildrequest_additions.append('MARK')
            self.db.buildrequests.fakeUnclaimBuildRequest(11)
            # note that at this point brid 20 is still unclaimed, but we do
            # not get a new notification about it
        d.addCallback(unclaim)
        d.addCallback(lambda _ : self.master.pollDatabaseBuildRequests())
        def check(_):
            self.assertEqual(self.gotten_buildrequest_additions, [
                dict(bsid=9, brid=11, buildername='eleventy'),
                'MARK',
                dict(bsid=9, brid=20, buildername='twenty'),
                'MARK',
                dict(bsid=9, brid=11, buildername='eleventy'),
            ])
        d.addCallback(check)
        return d




# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa

def upgrade(migrate_engine):

    metadata = sa.MetaData()
    metadata.bind = migrate_engine

    users_table = sa.Table('users', metadata, autoload=True)

    username = sa.Column('bb_username', sa.String(128))
    username.create(users_table)
    password = sa.Column('bb_password', sa.String(128))
    password.create(users_table)

    idx = sa.Index('users_bb_user', users_table.c.bb_username, unique=True)
    idx.create()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest
from buildbot.test.util import migration
import sqlalchemy as sa
from sqlalchemy.engine import reflection

class Migration(migration.MigrateTestMixin, unittest.TestCase):

    def setUp(self):
        return self.setUpMigrateTest()

    def tearDown(self):
        return self.tearDownMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        self.buildrequests = sa.Table('buildrequests', metadata,
            sa.Column('id', sa.Integer,  primary_key=True),
            sa.Column('buildsetid', sa.Integer, # foreign key removed
                nullable=False),
            sa.Column('buildername', sa.String(length=256), nullable=False),
            sa.Column('priority', sa.Integer, nullable=False,
                server_default=sa.DefaultClause("0")),
            sa.Column('complete', sa.Integer,
                server_default=sa.DefaultClause("0")),
            sa.Column('results', sa.SmallInteger),
            sa.Column('submitted_at', sa.Integer, nullable=False),
            sa.Column('complete_at', sa.Integer),
        )
        self.buildrequests.create(bind=conn)

        # these indices should already exist everywhere but on sqlite
        if conn.dialect.name != 'sqlite':
            idx = sa.Index('buildrequests_buildsetid',
                    self.buildrequests.c.buildsetid)
            idx.create()

            idx = sa.Index('buildrequests_buildername',
                    self.buildrequests.c.buildername)
            idx.create()

            idx = sa.Index('buildrequests_complete',
                    self.buildrequests.c.complete)
            idx.create()

    # tests

    def test_migrate(self):
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            insp = reflection.Inspector.from_engine(conn)
            indexes = insp.get_indexes('buildrequests')
            self.assertEqual(
                sorted([ i['name'] for i in indexes ]),
                sorted([
                    'buildrequests_buildername',
                    'buildrequests_buildsetid',
                    'buildrequests_complete',
                ]))


        return self.do_test_migration(15, 16, setup_thd, verify_thd)




from setuptools import setup

from bbwatcher import __version__

setup(
    name='Trac-BuildBot-Watcher',
    version=__version__,
    packages=['bbwatcher'],
    package_data={
        'bbwatcher': ['htdocs/*', 'templates/*.html'],
        },
    author='Randall Bosetti',
    description='A plugin to fetch/integrate status updates from the BuildBot XML-RPC interface',
    license='GPL',
    entry_points={
        'trac.plugins': 'bbwatcher.web_ui=bbwatcher.web_ui',
        }
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa

def upgrade(migrate_engine):
    metadata = sa.MetaData()
    metadata.bind = migrate_engine

    # re-include some of the relevant tables, as they were in version 3, since
    # sqlalchemy's reflection doesn't work very well for defaults.  These must
    # be complete table specifications as for some dialects sqlalchemy will
    # create a brand new, temporary table, and copy data over

    sa.Table("schedulers", metadata,
        sa.Column('schedulerid', sa.Integer, autoincrement=False,
            primary_key=True),
        sa.Column('name', sa.String(128), nullable=False),
        sa.Column('state', sa.String(1024), nullable=False),
        sa.Column('class_name', sa.String(128), nullable=False),
    )

    sa.Table('changes', metadata,
        sa.Column('changeid', sa.Integer, autoincrement=False,
            primary_key=True),
        sa.Column('author', sa.String(256), nullable=False),
        sa.Column('comments', sa.String(1024), nullable=False),
        sa.Column('is_dir', sa.SmallInteger, nullable=False),
        sa.Column('branch', sa.String(256)),
        sa.Column('revision', sa.String(256)),
        sa.Column('revlink', sa.String(256)),
        sa.Column('when_timestamp', sa.Integer, nullable=False),
        sa.Column('category', sa.String(256)),
        sa.Column('repository', sa.Text, nullable=False, server_default=''),
        sa.Column('project', sa.Text, nullable=False, server_default=''),
    )

    sa.Table('patches', metadata,
        sa.Column('id', sa.Integer,  primary_key=True),
        sa.Column('patchlevel', sa.Integer, nullable=False),
        sa.Column('patch_base64', sa.Text, nullable=False),
        sa.Column('subdir', sa.Text),
    )

    sa.Table('sourcestamps', metadata,
        sa.Column('id', sa.Integer, autoincrement=True, primary_key=True),
        sa.Column('branch', sa.String(256)),
        sa.Column('revision', sa.String(256)),
        sa.Column('patchid', sa.Integer, sa.ForeignKey('patches.id')),
        sa.Column('repository', sa.Text(length=None), nullable=False,
            server_default=''),
        sa.Column('project', sa.Text(length=None), nullable=False,
            server_default=''),
    )

    sa.Table('buildsets', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('external_idstring', sa.String(256)),
        sa.Column('reason', sa.String(256)),
        sa.Column('sourcestampid', sa.Integer,
            sa.ForeignKey('sourcestamps.id'), nullable=False),
        sa.Column('submitted_at', sa.Integer, nullable=False),
        sa.Column('complete', sa.SmallInteger, nullable=False,
            server_default=sa.DefaultClause("0")),
        sa.Column('complete_at', sa.Integer),
        sa.Column('results', sa.SmallInteger),
    )

    sa.Table('buildrequests', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('buildsetid', sa.Integer, sa.ForeignKey("buildsets.id"),
            nullable=False),
        sa.Column('buildername', sa.String(length=None), nullable=False),
        sa.Column('priority', sa.Integer, nullable=False,
            server_default=sa.DefaultClause("0")),
        sa.Column('claimed_at', sa.Integer,
            server_default=sa.DefaultClause("0")),
        sa.Column('claimed_by_name', sa.String(length=None)),
        sa.Column('claimed_by_incarnation', sa.String(length=None)),
        sa.Column('complete', sa.Integer,
            server_default=sa.DefaultClause("0")),
        sa.Column('results', sa.SmallInteger),
        sa.Column('submitted_at', sa.Integer, nullable=False),
        sa.Column('complete_at', sa.Integer),
    )

    sa.Table('builds', metadata,
        sa.Column('id', sa.Integer,  primary_key=True),
        sa.Column('number', sa.Integer, nullable=False),
        sa.Column('brid', sa.Integer, sa.ForeignKey('buildrequests.id'),
            nullable=False),
        sa.Column('start_time', sa.Integer, nullable=False),
        sa.Column('finish_time', sa.Integer),
    )

    to_autoinc = [ s.split(".") for s in
        "schedulers.schedulerid",
        "builds.id",
        "changes.changeid",
        "buildrequests.id",
        "buildsets.id",
        "patches.id",
        "sourcestamps.id",
    ]

    # It seems that SQLAlchemy's ALTER TABLE doesn't work when migrating from
    # INTEGER to PostgreSQL's SERIAL data type (which is just pseudo data type
    # for INTEGER with SEQUENCE), so we have to work-around this with raw SQL.
    if migrate_engine.dialect.name in ('postgres', 'postgresql'):
        for table_name, col_name in to_autoinc:
            migrate_engine.execute("CREATE SEQUENCE %s_%s_seq"
                                   % (table_name, col_name))
            migrate_engine.execute("ALTER TABLE %s ALTER COLUMN %s SET DEFAULT nextval('%s_%s_seq'::regclass)"
                                   % (table_name, col_name, table_name, col_name))
            migrate_engine.execute("ALTER SEQUENCE %s_%s_seq OWNED BY %s.%s"
                                   % (table_name, col_name, table_name, col_name))
    else:
        for table_name, col_name in to_autoinc:
            table = metadata.tables[table_name]
            col = table.c[col_name]
            col.alter(autoincrement=True)


    # also drop the changes_nextid table here (which really should have been a
    # sequence..)
    table = sa.Table('changes_nextid', metadata, autoload=True)
    table.drop()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from twisted.trial import unittest
from buildbot.db import users
from buildbot.test.util import connector_component
from buildbot.test.fake import fakedb

class TestUsersConnectorComponent(connector_component.ConnectorComponentMixin,
                                 unittest.TestCase):

    def setUp(self):
        d = self.setUpConnectorComponent(
                table_names=['users', 'users_info', 'changes', 'change_users'])
        def finish_setup(_):
            self.db.users = users.UsersConnectorComponent(self.db)
        d.addCallback(finish_setup)
        return d

    def tearDown(self):
        return self.tearDownConnectorComponent()

    # sample user data

    user1_rows = [
        fakedb.User(uid=1, identifier='soap'),
        fakedb.UserInfo(uid=1, attr_type='IPv9', attr_data='0578cc6.8db024'),
    ]

    user2_rows = [
        fakedb.User(uid=2, identifier='lye'),
        fakedb.UserInfo(uid=2, attr_type='git',
                        attr_data='Tyler Durden <tyler@mayhem.net>'),
        fakedb.UserInfo(uid=2, attr_type='irc', attr_data='durden')
    ]

    user3_rows = [
        fakedb.User(uid=3, identifier='marla', bb_username='marla',
                    bb_password='cancer')
    ]

    user1_dict = {
        'uid': 1,
        'identifier': u'soap',
        'bb_username': None,
        'bb_password': None,
        'IPv9': u'0578cc6.8db024',
    }

    user2_dict = {
        'uid': 2,
        'identifier': u'lye',
        'bb_username': None,
        'bb_password': None,
        'irc': u'durden',
        'git': u'Tyler Durden <tyler@mayhem.net>'
    }

    user3_dict = {
        'uid': 3,
        'identifier': u'marla',
        'bb_username': u'marla',
        'bb_password': u'cancer',
    }

    # tests

    def test_addUser_new(self):
        d = self.db.users.findUserByAttr(identifier='soap',
                                  attr_type='subspace_net_handle',
                                  attr_data='Durden0924')
        def check_user(uid):
            def thd(conn):
                users_tbl = self.db.model.users
                users_info_tbl = self.db.model.users_info
                users = conn.execute(users_tbl.select()).fetchall()
                infos = conn.execute(users_info_tbl.select()).fetchall()
                self.assertEqual(len(users), 1)
                self.assertEqual(users[0].uid, uid)
                self.assertEqual(users[0].identifier, 'soap')
                self.assertEqual(len(infos), 1)
                self.assertEqual(infos[0].uid, uid)
                self.assertEqual(infos[0].attr_type, 'subspace_net_handle')
                self.assertEqual(infos[0].attr_data, 'Durden0924')
            return self.db.pool.do(thd)
        d.addCallback(check_user)
        return d

    def test_addUser_existing(self):
        d = self.insertTestData(self.user1_rows)
        d.addCallback(lambda _ : self.db.users.findUserByAttr(
                                  identifier='soapy',
                                  attr_type='IPv9',
                                  attr_data='0578cc6.8db024'))
        def check_user(uid):
            self.assertEqual(uid, 1)
            def thd(conn):
                users_tbl = self.db.model.users
                users_info_tbl = self.db.model.users_info
                users = conn.execute(users_tbl.select()).fetchall()
                infos = conn.execute(users_info_tbl.select()).fetchall()
                self.assertEqual(len(users), 1)
                self.assertEqual(users[0].uid, uid)
                self.assertEqual(users[0].identifier, 'soap') # not changed!
                self.assertEqual(len(infos), 1)
                self.assertEqual(infos[0].uid, uid)
                self.assertEqual(infos[0].attr_type, 'IPv9')
                self.assertEqual(infos[0].attr_data, '0578cc6.8db024')
            return self.db.pool.do(thd)
        d.addCallback(check_user)
        return d

    def test_findUser_existing(self):
        d = self.insertTestData(self.user1_rows)
        d = self.insertTestData(self.user2_rows)
        d = self.insertTestData(self.user3_rows)
        d.addCallback(lambda _ : self.db.users.findUserByAttr(
                                  identifier='lye',
                                  attr_type='git',
                                  attr_data='Tyler Durden <tyler@mayhem.net>'))
        def check_user(uid):
            self.assertEqual(uid, 2)
            def thd(conn):
                users_tbl = self.db.model.users
                users_info_tbl = self.db.model.users_info
                users = conn.execute(users_tbl.select()).fetchall()
                infos = conn.execute(users_info_tbl.select()).fetchall()
                self.assertEqual((
                    sorted([ tuple(u) for u in users]),
                    sorted([ tuple(i) for i in infos])
                ), (
                    [
                        (1L, u'soap', None, None),
                        (2L, u'lye', None, None),
                        (3L, u'marla', u'marla', u'cancer'),
                    ], [
                        (1L, u'IPv9', u'0578cc6.8db024'),
                        (2L, u'git', u'Tyler Durden <tyler@mayhem.net>'),
                        (2L, u'irc', u'durden')
                    ]))
            return self.db.pool.do(thd)
        d.addCallback(check_user)
        return d

    def test_addUser_race(self):
        def race_thd(conn):
            # note that this assumes that both inserts can happen "at once".
            # This is the case for DB engines that support transactions, but
            # not for MySQL.  so this test does not detect the potential MySQL
            # failure, which will generally result in a spurious failure.
            conn.execute(self.db.model.users.insert(),
                    uid=99, identifier='soap')
            conn.execute(self.db.model.users_info.insert(),
                    uid=99, attr_type='subspace_net_handle',
                    attr_data='Durden0924')
        d = self.db.users.findUserByAttr(identifier='soap',
                                  attr_type='subspace_net_handle',
                                  attr_data='Durden0924',
                                  _race_hook=race_thd)
        def check_user(uid):
            self.assertEqual(uid, 99)
            def thd(conn):
                users_tbl = self.db.model.users
                users_info_tbl = self.db.model.users_info
                users = conn.execute(users_tbl.select()).fetchall()
                infos = conn.execute(users_info_tbl.select()).fetchall()
                self.assertEqual(len(users), 1)
                self.assertEqual(users[0].uid, uid)
                self.assertEqual(users[0].identifier, 'soap')
                self.assertEqual(len(infos), 1)
                self.assertEqual(infos[0].uid, uid)
                self.assertEqual(infos[0].attr_type, 'subspace_net_handle')
                self.assertEqual(infos[0].attr_data, 'Durden0924')
            return self.db.pool.do(thd)
        d.addCallback(check_user)
        return d

    def test_addUser_existing_identifier(self):
        d = self.insertTestData(self.user1_rows)
        d.addCallback(lambda _ : self.db.users.findUserByAttr(
                                  identifier='soap',
                                  attr_type='telepathIO(tm)',
                                  attr_data='hmm,lye'))
        def cb(_):
            self.fail("shouldn't get here")
        def eb(f):
            f.trap(sa.exc.IntegrityError, sa.exc.ProgrammingError)
            pass # expected
        d.addCallbacks(cb, eb)
        return d

    def test_getUser(self):
        d = self.insertTestData(self.user1_rows)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict, self.user1_dict)
        d.addCallback(check1)
        return d

    def test_getUser_bb(self):
        d = self.insertTestData(self.user3_rows)
        def get3(_):
            return self.db.users.getUser(3)
        d.addCallback(get3)
        def check3(usdict):
            self.assertEqual(usdict, self.user3_dict)
        d.addCallback(check3)
        return d

    def test_getUser_multi_attr(self):
        d = self.insertTestData(self.user2_rows)
        def get1(_):
            return self.db.users.getUser(2)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict, self.user2_dict)
        d.addCallback(check1)
        return d

    def test_getUser_no_match(self):
        d = self.insertTestData(self.user1_rows)
        def get3(_):
            return self.db.users.getUser(3)
        d.addCallback(get3)
        def check3(none):
            self.assertEqual(none, None)
        d.addCallback(check3)
        return d

    def test_getUsers_none(self):
        d = self.db.users.getUsers()
        def check(res):
            self.assertEqual(res, [])
        d.addCallback(check)
        return d

    def test_getUsers(self):
        d = self.insertTestData(self.user1_rows)
        def get(_):
            return self.db.users.getUsers()
        d.addCallback(get)
        def check(res):
            self.assertEqual(res, [dict(uid=1, identifier='soap')])
        d.addCallback(check)
        return d

    def test_getUsers_multiple(self):
        d = self.insertTestData(self.user1_rows + self.user2_rows)
        def get(_):
            return self.db.users.getUsers()
        d.addCallback(get)
        def check(res):
            self.assertEqual(res, [dict(uid=1, identifier='soap'),
                                   dict(uid=2, identifier='lye')])
        d.addCallback(check)
        return d

    def test_getUserByUsername(self):
        d = self.insertTestData(self.user3_rows)
        def get3(_):
            return self.db.users.getUserByUsername("marla")
        d.addCallback(get3)
        def check3(res):
            self.assertEqual(res, self.user3_dict)
        d.addCallback(check3)
        return d

    def test_getUserByUsername_no_match(self):
        d = self.insertTestData(self.user3_rows)
        def get3(_):
            return self.db.users.getUserByUsername("tyler")
        d.addCallback(get3)
        def check3(none):
            self.assertEqual(none, None)
        d.addCallback(check3)
        return d

    def test_updateUser_existing_type(self):
        d = self.insertTestData(self.user1_rows)
        def update1(_):
            return self.db.users.updateUser(
                uid=1, attr_type='IPv9', attr_data='abcd.1234')
        d.addCallback(update1)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['IPv9'], 'abcd.1234')
            self.assertEqual(usdict['identifier'], 'soap') # no change
        d.addCallback(check1)
        return d

    def test_updateUser_new_type(self):
        d = self.insertTestData(self.user1_rows)
        def update1(_):
            return self.db.users.updateUser(
                uid=1, attr_type='IPv4', attr_data='123.134.156.167')
        d.addCallback(update1)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['IPv4'], '123.134.156.167')
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
            self.assertEqual(usdict['identifier'], 'soap') # no change
        d.addCallback(check1)
        return d

    def test_updateUser_identifier(self):
        d = self.insertTestData(self.user1_rows)
        def update1(_):
            return self.db.users.updateUser(
                uid=1, identifier='lye')
        d.addCallback(update1)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['identifier'], 'lye')
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
        d.addCallback(check1)
        return d

    def test_updateUser_bb(self):
        d = self.insertTestData(self.user3_rows)
        def update3(_):
            return self.db.users.updateUser(
                uid=3, bb_username='boss', bb_password='fired')
        d.addCallback(update3)
        def get3(_):
            return self.db.users.getUser(3)
        d.addCallback(get3)
        def check3(usdict):
            self.assertEqual(usdict['bb_username'], 'boss')
            self.assertEqual(usdict['bb_password'], 'fired')
            self.assertEqual(usdict['identifier'], 'marla') # no change
        d.addCallback(check3)
        return d

    def test_updateUser_all(self):
        d = self.insertTestData(self.user1_rows)
        def update1(_):
            return self.db.users.updateUser(
                uid=1, identifier='lye', bb_username='marla',
                bb_password='cancer', attr_type='IPv4', attr_data='123.134.156.167')
        d.addCallback(update1)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['identifier'], 'lye')
            self.assertEqual(usdict['bb_username'], 'marla')
            self.assertEqual(usdict['bb_password'], 'cancer')
            self.assertEqual(usdict['IPv4'], '123.134.156.167')
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
        d.addCallback(check1)
        return d

    def test_updateUser_race(self):
        # called from the db thread, this opens a *new* connection (to avoid
        # the existing transaction) and executes a conflicting insert in that
        # connection.  This will cause the insert in the db method to fail, and
        # the data in this insert (8.8.8.8) will appear below.
        def race_thd(conn):
            conn = self.db.pool.engine.connect()
            conn.execute(self.db.model.users_info.insert(),
                    uid=1, attr_type='IPv4',
                    attr_data='8.8.8.8')

        d = self.insertTestData(self.user1_rows)
        def update1(_):
            return self.db.users.updateUser(
                uid=1, attr_type='IPv4', attr_data='123.134.156.167',
                _race_hook=race_thd)
        d.addCallback(update1)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['identifier'], 'soap')
            self.assertEqual(usdict['IPv4'], '8.8.8.8')
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
        d.addCallback(check1)
        return d

    def test_update_NoMatch_identifier(self):
        d = self.insertTestData(self.user1_rows)
        def update3(_):
            return self.db.users.updateUser(
                uid=3, identifier='abcd')
        d.addCallback(update3)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['identifier'], 'soap') # no change
        d.addCallback(check1)
        return d

    def test_update_NoMatch_attribute(self):
        d = self.insertTestData(self.user1_rows)
        def update3(_):
            return self.db.users.updateUser(
                uid=3, attr_type='abcd', attr_data='efgh')
        d.addCallback(update3)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
        d.addCallback(check1)
        return d

    def test_update_NoMatch_bb(self):
        d = self.insertTestData(self.user1_rows)
        def update3(_):
            return self.db.users.updateUser(
                uid=3, attr_type='marla', attr_data='cancer')
        d.addCallback(update3)
        def get1(_):
            return self.db.users.getUser(1)
        d.addCallback(get1)
        def check1(usdict):
            self.assertEqual(usdict['IPv9'], '0578cc6.8db024') # no change
        d.addCallback(check1)
        return d

    def test_removeUser_uid(self):
        d = self.insertTestData(self.user1_rows)
        def remove1(_):
            return self.db.users.removeUser(1)
        d.addCallback(remove1)
        def check1(_):
            def thd(conn):
                r = conn.execute(self.db.model.users.select())
                r = r.fetchall()
                self.assertEqual(len(r), 0)
            return self.db.pool.do(thd)
        d.addCallback(check1)
        return d

    def test_removeNoMatch(self):
        d = self.insertTestData(self.user1_rows)
        def check(_):
            return self.db.users.removeUser(uid=3)
        d.addCallback(check)
        return d

    def test_identifierToUid_NoMatch(self):
        d = self.db.users.identifierToUid(identifier="soap")
        def check(res):
            self.assertEqual(res, None)
        d.addCallback(check)
        return d

    def test_identifierToUid_match(self):
        d = self.insertTestData(self.user1_rows)
        def ident2uid(_):
            return self.db.users.identifierToUid(identifier="soap")
        d.addCallback(ident2uid)
        def check(res):
            self.assertEqual(res, 1)
        d.addCallback(check)
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import mock
from twisted.python import failure, log
from twisted.trial import unittest
from twisted.internet import defer, reactor, task
from twisted.spread import pb
from buildbot.test.util import compat
from buildbot.process import botmaster

# TODO: should do this with more of the Twisted machinery intact - maybe in a
# separate integration test?

class FakeAbstractBuildSlave(object):

    def __init__(self, slave, name, isConnected=True, reactor=reactor):
        self.slavename = name
        self.slave = slave
        self.isConnectedResult = isConnected
        self.reactor = reactor
        self.call_on_detach = lambda : None

        # set up for loseConnection to cause the slave to detach, but not
        # immediately
        def tport_loseConnection():
            self.isConnectedResult = False
            self.call_on_detach()
            self.call_on_detach = None
        self.slave.broker.transport.loseConnection = (lambda :
                self.reactor.callLater(0, tport_loseConnection))

    def subscribeToDetach(self, callback):
        self.call_on_detach = callback

    def isConnected(self):
        return self.isConnectedResult


class FakeRemoteBuildSlave(object):

    def __init__(self, name, callRemoteFailure=False,
                             callRemoteException=False,
                             callRemoteHang=0,
                             reactor=reactor):
        self.name = name
        self.callRemoteFailure = callRemoteFailure
        self.callRemoteException = callRemoteException
        self.callRemoteHang = callRemoteHang
        self.reactor = reactor

        self.broker = mock.Mock()
        self.broker.transport.getPeer = lambda : "<peer %s>" % name

    def _makePingResult(self):
        if self.callRemoteException:
            exc = self.callRemoteException()
            log.msg(" -> exception %r" % (exc,))
            raise exc
        if self.callRemoteFailure:
            f = defer.fail(self.callRemoteFailure())
            log.msg(" -> failure %r" % (f,))
            return f
        return defer.succeed(None)

    def callRemote(self, meth, *args, **kwargs):
        assert meth == "print"
        log.msg("%r.callRemote('print', %r)" % (self, args[0],))
        # if we're asked to hang, then set up to fire the deferred later
        if self.callRemoteHang:
            log.msg(" -> hang for %d s" % (self.callRemoteHang,))
            d = defer.Deferred()
            self.reactor.callLater(self.callRemoteHang, d.callback, None)
            def hangdone(_):
                log.msg("%r.callRemote hang finished" % (self,))
                return self._makePingResult()
            d.addCallback(hangdone)
            self.callRemote_d = d
        # otherwise, return a fired deferred
        else:
            self.callRemote_d = self._makePingResult()
        return self.callRemote_d

    def __repr__(self):
        return "<FakeRemoteBuildSlave %s>" % (self.name,)


class DuplicateSlaveArbitrator(unittest.TestCase):

    def makeDeadReferenceError(self):
        return pb.DeadReferenceError("Calling Stale Broker (fake exception)")

    def makeRuntimeError(self):
        return RuntimeError("oh noes!")

    def makePBConnectionLostFailure(self):
        return failure.Failure(pb.PBConnectionLost("gone"))

    def test_old_slave_present(self):
        old_remote = FakeRemoteBuildSlave("old")
        new_remote = FakeRemoteBuildSlave("new")
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave")
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            self.fail("shouldn't get here")
        def failed(f):
            f.trap(RuntimeError) # expected error
        d.addCallbacks(got_persp, failed)
        return d

    def test_old_slave_absent_deadref_exc(self):
        old_remote = FakeRemoteBuildSlave("old",
                callRemoteException=self.makeDeadReferenceError)
        new_remote = FakeRemoteBuildSlave("new")
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave")
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            self.assertIdentical(bs, buildslave)
        d.addCallback(got_persp)
        return d

    def test_old_slave_absent_connlost_failure(self):
        old_remote = FakeRemoteBuildSlave("old",
                callRemoteFailure=self.makePBConnectionLostFailure)
        new_remote = FakeRemoteBuildSlave("new")
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave")
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            self.assertIdentical(bs, buildslave)
        d.addCallback(got_persp)
        return d

    @compat.usesFlushLoggedErrors
    def test_old_slave_absent_unexpected_exc(self):
        old_remote = FakeRemoteBuildSlave("old",
                callRemoteException=self.makeRuntimeError)
        new_remote = FakeRemoteBuildSlave("new")
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave")
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            # getPerspective has returned successfully:
            self.assertIdentical(bs, buildslave)
            # and the unexpected RuntimeError was logged:
            self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)
        d.addCallback(got_persp)
        return d

    def do_test_old_slave_absent_timeout(self, callRemoteException=None):
        clock = task.Clock()
        PING_TIMEOUT = botmaster.DuplicateSlaveArbitrator.PING_TIMEOUT

        old_remote = FakeRemoteBuildSlave("old", reactor=clock,
                callRemoteHang=PING_TIMEOUT+1,
                callRemoteException=callRemoteException)
        new_remote = FakeRemoteBuildSlave("new")
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave",
                reactor=clock)
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        arb._reactor = clock
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            self.assertIdentical(bs, buildslave)
        d.addCallback(got_persp)

        # show the passage of time for 2s more than the PING_TIMEOUT, to allow
        # the old callRemote to return eventually
        clock.pump([.1] * 10 * (PING_TIMEOUT+4))

        # check that the timed-out call eventually returned (and was ignored,
        # even if there was an exception)
        self.failUnless(old_remote.callRemote_d.called)

        return d

    def test_old_slave_absent_timeout(self):
        return self.do_test_old_slave_absent_timeout()

    def test_old_slave_absent_timeout_exc(self):
        return self.do_test_old_slave_absent_timeout(
                callRemoteException=self.makeRuntimeError)

    @compat.usesFlushLoggedErrors
    def test_new_slave_ping_error(self):
        old_remote = FakeRemoteBuildSlave("old")
        new_remote = FakeRemoteBuildSlave("new",
                callRemoteException=self.makeRuntimeError)
        buildslave = FakeAbstractBuildSlave(old_remote, name="testslave")
        arb = botmaster.DuplicateSlaveArbitrator(buildslave)
        d = arb.getPerspective(new_remote, "testslave")
        def got_persp(bs):
            self.fail("shouldn't get here")
        def failed(f):
            pass #f.trap(RuntimeError) # expected error
        d.addCallbacks(got_persp, failed)
        return d




# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from twisted.trial import unittest
from buildbot.test.util import migration

class Migration(migration.MigrateTestMixin, unittest.TestCase):

    def setUp(self):
        return self.setUpMigrateTest()

    def tearDown(self):
        return self.tearDownMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        changes = sa.Table('changes', metadata,
            sa.Column('changeid', sa.Integer,  primary_key=True),
            # the rest is unimportant
        )
        changes.create()

        buildsets = sa.Table('buildsets', metadata,
            sa.Column('id', sa.Integer,  primary_key=True),
            # the rest is unimportant
        )
        buildsets.create()

        self.schedulers = sa.Table("schedulers", metadata,
            sa.Column('schedulerid', sa.Integer, primary_key=True),
            sa.Column('name', sa.String(128), nullable=False),
            sa.Column('class_name', sa.String(128), nullable=False),
        )
        self.schedulers.create(bind=conn)
        sa.Index('name_and_class', self.schedulers.c.name,
                self.schedulers.c.class_name).create()

        self.scheduler_changes = sa.Table('scheduler_changes', metadata,
            sa.Column('schedulerid', sa.Integer,
                sa.ForeignKey('schedulers.schedulerid')),
            sa.Column('changeid', sa.Integer,
                sa.ForeignKey('changes.changeid')),
            sa.Column('important', sa.SmallInteger),
        )
        self.scheduler_changes.create()
        sa.Index('scheduler_changes_schedulerid',
                self.scheduler_changes.c.schedulerid).create()
        sa.Index('scheduler_changes_changeid',
                self.scheduler_changes.c.changeid).create()
        sa.Index('scheduler_changes_unique',
                self.scheduler_changes.c.schedulerid,
                self.scheduler_changes.c.changeid, unique=True).create()

        self.scheduler_upstream_buildsets = sa.Table(
                    'scheduler_upstream_buildsets', metadata,
            sa.Column('buildsetid', sa.Integer, sa.ForeignKey('buildsets.id')),
            sa.Column('schedulerid', sa.Integer,
                sa.ForeignKey('schedulers.schedulerid')),
        )
        self.scheduler_upstream_buildsets.create()

        sa.Index('scheduler_upstream_buildsets_buildsetid',
                self.scheduler_upstream_buildsets.c.buildsetid).create()
        sa.Index('scheduler_upstream_buildsets_schedulerid',
                self.scheduler_upstream_buildsets.c.schedulerid).create()

        self.objects = sa.Table("objects", metadata,
            sa.Column("id", sa.Integer, primary_key=True),
            sa.Column('name', sa.String(128), nullable=False),
            sa.Column('class_name', sa.String(128), nullable=False),
        )
        self.objects.create(bind=conn)

        sa.Index('object_identity', self.objects.c.name,
                self.objects.c.class_name, unique=True).create()

    # tests

    def test_update(self):
        # this upgrade script really just drops a bunch of tables, so
        # there's not much to test!
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            metadata = sa.MetaData()
            metadata.bind = conn

            # these tables are gone
            for tbl in 'schedulers', 'scheduler_upstream_buildsets':
                try:
                    conn.execute("select * from %s" % tbl)
                except:
                    pass
                else:
                    self.fail("%s table still exists" % tbl)

            # but scheduler_changes is not
            s_c_tbl = sa.Table("scheduler_changes", metadata,
                    autoload=True)
            q = sa.select(
               [ s_c_tbl.c.objectid, s_c_tbl.c.changeid, s_c_tbl.c.important ])
            self.assertEqual(conn.execute(q).fetchall(), [])

        return self.do_test_migration(18, 19, setup_thd, verify_thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest
from buildbot.db import schedulers
from buildbot.test.util import connector_component
from buildbot.test.fake import fakedb

class TestSchedulersConnectorComponent(
            connector_component.ConnectorComponentMixin,
            unittest.TestCase):

    def setUp(self):
        d = self.setUpConnectorComponent(
            table_names=['changes', 'objects', 'scheduler_changes' ])

        def finish_setup(_):
            self.db.schedulers = \
                    schedulers.SchedulersConnectorComponent(self.db)
        d.addCallback(finish_setup)

        return d

    def tearDown(self):
        return self.tearDownConnectorComponent()

    def checkScheduler(self, objectid, name, class_name):
        def thd(conn):
            q = self.db.model.schedulers.select(
                whereclause=(self.db.model.schedulers.c.objectid == objectid))
            for row in conn.execute(q):
                self.assertEqual([ row.objectid, row.name, row.class_name],
                                 [ objectid, name, class_name])
        return self.db.pool.do(thd)

    # test data

    change3 = fakedb.Change(changeid=3)
    change4 = fakedb.Change(changeid=4)
    change5 = fakedb.Change(changeid=5)
    change6 = fakedb.Change(changeid=6, branch='sql')

    scheduler24 = fakedb.Object(id=24)

    def addClassifications(self, _, objectid, *classifications):
        def thd(conn):
            q = self.db.model.scheduler_changes.insert()
            conn.execute(q, [
                dict(changeid=c[0], objectid=objectid, important=c[1])
                for c in classifications ])
        return self.db.pool.do(thd)

    # tests
    def test_classifyChanges(self):
        d = self.insertTestData([ self.change3, self.change4,
                                  self.scheduler24 ])
        d.addCallback(lambda _ :
                self.db.schedulers.classifyChanges(24,
                    { 3 : False, 4: True }))
        def check(_):
            def thd(conn):
                sch_chgs_tbl = self.db.model.scheduler_changes
                q = sch_chgs_tbl.select(order_by=sch_chgs_tbl.c.changeid)
                r = conn.execute(q)
                rows = [ (row.objectid, row.changeid, row.important)
                         for row in r.fetchall() ]
                self.assertEqual(rows, [ (24, 3, 0), (24, 4, 1) ])
            return self.db.pool.do(thd)
        d.addCallback(check)
        return d

    def test_classifyChanges_again(self):
        # test reclassifying changes, which may happen during some timing
        # conditions
        d = self.insertTestData([
            self.change3,
            self.scheduler24,
            fakedb.SchedulerChange(objectid=24, changeid=3, important=0),
        ])
        d.addCallback(lambda _ :
                self.db.schedulers.classifyChanges(24, { 3 : True }))
        def check(_):
            def thd(conn):
                sch_chgs_tbl = self.db.model.scheduler_changes
                q = sch_chgs_tbl.select(order_by=sch_chgs_tbl.c.changeid)
                r = conn.execute(q)
                rows = [ (row.objectid, row.changeid, row.important)
                         for row in r.fetchall() ]
                self.assertEqual(rows, [ (24, 3, 1) ])
            return self.db.pool.do(thd)
        d.addCallback(check)
        return d

    def test_flushChangeClassifications(self):
        d = self.insertTestData([ self.change3, self.change4,
                                  self.change5, self.scheduler24 ])
        d.addCallback(self.addClassifications, 24,
                (3, 1), (4, 0), (5, 1))
        d.addCallback(lambda _ :
            self.db.schedulers.flushChangeClassifications(24))
        def check(_):
            def thd(conn):
                q = self.db.model.scheduler_changes.select()
                rows = conn.execute(q).fetchall()
                self.assertEqual(rows, [])
            return self.db.pool.do(thd)
        d.addCallback(check)
        return d

    def test_flushChangeClassifications_less_than(self):
        d = self.insertTestData([ self.change3, self.change4,
                                  self.change5, self.scheduler24 ])
        d.addCallback(self.addClassifications, 24,
                (3, 1), (4, 0), (5, 1))
        d.addCallback(lambda _ :
            self.db.schedulers.flushChangeClassifications(24, less_than=5))
        def check(_):
            def thd(conn):
                q = self.db.model.scheduler_changes.select()
                rows = conn.execute(q).fetchall()
                self.assertEqual([ (r.changeid, r.important) for r in rows],
                                 [ (5, 1) ])
            return self.db.pool.do(thd)
        d.addCallback(check)
        return d

    def test_getChangeClassifications(self):
        d = self.insertTestData([ self.change3, self.change4, self.change5,
                                  self.change6, self.scheduler24 ])
        d.addCallback(self.addClassifications, 24,
                (3, 1), (4, 0), (5, 1), (6, 1))
        d.addCallback(lambda _ :
            self.db.schedulers.getChangeClassifications(24))
        def check(cls):
            self.assertEqual(cls, { 3 : True, 4 : False, 5 : True, 6: True })
        d.addCallback(check)
        return d

    def test_getChangeClassifications_branch(self):
        d = self.insertTestData([ self.change3, self.change4, self.change5,
                                  self.change6, self.scheduler24 ])
        d.addCallback(self.addClassifications, 24,
                (3, 1), (4, 0), (5, 1), (6, 1))
        d.addCallback(lambda _ :
            self.db.schedulers.getChangeClassifications(24, branch='sql'))
        def check(cls):
            self.assertEqual(cls, { 6 : True })
        d.addCallback(check)
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest
from twisted.internet import defer, reactor, task
from buildbot.test.util import changesource, compat
from buildbot.changes import base

class TestPollingChangeSource(changesource.ChangeSourceMixin, unittest.TestCase):
    class Subclass(base.PollingChangeSource):
        pass

    def setUp(self):
        # patch in a Clock so we can manipulate the reactor's time
        self.clock = task.Clock()
        self.patch(reactor, 'callLater', self.clock.callLater)
        self.patch(reactor, 'seconds', self.clock.seconds)

        d = self.setUpChangeSource()
        def create_changesource(_):
            self.attachChangeSource(self.Subclass())
        d.addCallback(create_changesource)
        return d

    def tearDown(self):
        return self.tearDownChangeSource()

    def runClockFor(self, _, secs):
        self.clock.pump([1.0] * secs)

    def test_loop_loops(self):
        # track when poll() gets called
        loops = []
        self.changesource.poll = \
                lambda : loops.append(self.clock.seconds())

        self.changesource.pollInterval = 5
        self.startChangeSource()

        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)
        def check(_):
            # note that it does *not* poll at time 0
            self.assertEqual(loops, [5.0, 10.0])
        d.addCallback(check)
        reactor.callWhenRunning(d.callback, None)
        return d

    @compat.usesFlushLoggedErrors
    def test_loop_exception(self):
        # track when poll() gets called
        loops = []
        def poll():
            loops.append(self.clock.seconds())
            raise RuntimeError("oh noes")
        self.changesource.poll = poll

        self.changesource.pollInterval = 5
        self.startChangeSource()

        d = defer.Deferred()
        d.addCallback(self.runClockFor, 12)
        def check(_):
            # note that it keeps looping after error
            self.assertEqual(loops, [5.0, 10.0])
            self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 2)
        d.addCallback(check)
        reactor.callWhenRunning(d.callback, None)
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.python import log, reflect
from twisted.python.failure import Failure
from twisted.internet import defer, reactor
from twisted.spread import pb
from twisted.application import service

from buildbot.process.builder import Builder
from buildbot import interfaces, locks, config, util
from buildbot.process import metrics

class BotMaster(config.ReconfigurableServiceMixin, service.MultiService):

    """This is the master-side service which manages remote buildbot slaves.
    It provides them with BuildSlaves, and distributes build requests to
    them."""

    debug = 0

    def __init__(self, master):
        service.MultiService.__init__(self)
        self.setName("botmaster")
        self.master = master

        self.builders = {}
        self.builderNames = []
        # builders maps Builder names to instances of bb.p.builder.Builder,
        # which is the master-side object that defines and controls a build.

        # self.slaves contains a ready BuildSlave instance for each
        # potential buildslave, i.e. all the ones listed in the config file.
        # If the slave is connected, self.slaves[slavename].slave will
        # contain a RemoteReference to their Bot instance. If it is not
        # connected, that attribute will hold None.
        self.slaves = {} # maps slavename to BuildSlave
        self.watchers = {}

        # self.locks holds the real Lock instances
        self.locks = {}

        # self.mergeRequests is the callable override for merging build
        # requests
        self.mergeRequests = None

        self.shuttingDown = False

        self.lastSlavePortnum = None

        # subscription to new build requests
        self.buildrequest_sub = None

        # a distributor for incoming build requests; see below
        self.brd = BuildRequestDistributor(self)
        self.brd.setServiceParent(self)

    def cleanShutdown(self, _reactor=reactor):
        """Shut down the entire process, once all currently-running builds are
        complete."""
        if self.shuttingDown:
            return
        log.msg("Initiating clean shutdown")
        self.shuttingDown = True

        # first, stop the distributor; this will finish any ongoing scheduling
        # operations before firing
        d = self.brd.stopService()

        # then wait for all builds to finish
        def wait(_):
            l = []
            for builder in self.builders.values():
                for build in builder.builder_status.getCurrentBuilds():
                    l.append(build.waitUntilFinished())
            if len(l) == 0:
                log.msg("No running jobs, starting shutdown immediately")
            else:
                log.msg("Waiting for %i build(s) to finish" % len(l))
                return defer.DeferredList(l)
        d.addCallback(wait)

        # Finally, shut the whole process down
        def shutdown(ign):
            # Double check that we're still supposed to be shutting down
            # The shutdown may have been cancelled!
            if self.shuttingDown:
                # Check that there really aren't any running builds
                for builder in self.builders.values():
                    n = len(builder.builder_status.getCurrentBuilds())
                    if n > 0:
                        log.msg("Not shutting down, builder %s has %i builds running" % (builder, n))
                        log.msg("Trying shutdown sequence again")
                        self.shuttingDown = False
                        self.cleanShutdown()
                        return
                log.msg("Stopping reactor")
                _reactor.stop()
            else:
                self.brd.startService()
        d.addCallback(shutdown)
        d.addErrback(log.err, 'while processing cleanShutdown')

    def cancelCleanShutdown(self):
        """Cancel a clean shutdown that is already in progress, if any"""
        if not self.shuttingDown:
            return
        log.msg("Cancelling clean shutdown")
        self.shuttingDown = False

    @metrics.countMethod('BotMaster.slaveLost()')
    def slaveLost(self, bot):
        metrics.MetricCountEvent.log("BotMaster.attached_slaves", -1)
        for name, b in self.builders.items():
            if bot.slavename in b.config.slavenames:
                b.detached(bot)

    @metrics.countMethod('BotMaster.getBuildersForSlave()')
    def getBuildersForSlave(self, slavename):
        return [ b for b in self.builders.values()
                 if slavename in b.config.slavenames ]

    def getBuildernames(self):
        return self.builderNames

    def getBuilders(self):
        return self.builders.values()

    def startService(self):
        def buildRequestAdded(notif):
            self.maybeStartBuildsForBuilder(notif['buildername'])
        self.buildrequest_sub = \
            self.master.subscribeToBuildRequests(buildRequestAdded)
        service.MultiService.startService(self)

    @defer.deferredGenerator
    def reconfigService(self, new_config):
        timer = metrics.Timer("BotMaster.reconfigService")
        timer.start()

        # reconfigure slaves
        wfd = defer.waitForDeferred(
            self.reconfigServiceSlaves(new_config))
        yield wfd
        wfd.getResult()

        # reconfigure builders
        wfd = defer.waitForDeferred(
            self.reconfigServiceBuilders(new_config))
        yield wfd
        wfd.getResult()

        wfd = defer.waitForDeferred(
            config.ReconfigurableServiceMixin.reconfigService(self,
                                                    new_config))
        yield wfd
        wfd.getResult()

        # try to start a build for every builder; this is necessary at master
        # startup, and a good idea in any other case
        self.maybeStartBuildsForAllBuilders()

        timer.stop()


    @defer.deferredGenerator
    def reconfigServiceSlaves(self, new_config):

        timer = metrics.Timer("BotMaster.reconfigServiceSlaves")
        timer.start()

        # arrange slaves by name
        old_by_name = dict([ (s.slavename, s)
                            for s in list(self)
                            if interfaces.IBuildSlave.providedBy(s) ])
        old_set = set(old_by_name.iterkeys())
        new_by_name = dict([ (s.slavename, s)
                            for s in new_config.slaves ])
        new_set = set(new_by_name.iterkeys())

        # calculate new slaves, by name, and removed slaves
        removed_names, added_names = util.diffSets(old_set, new_set)

        # find any slaves for which the fully qualified class name has
        # changed, and treat those as an add and remove
        for n in old_set & new_set:
            old = old_by_name[n]
            new = new_by_name[n]
            # detect changed class name
            if reflect.qual(old.__class__) != reflect.qual(new.__class__):
                removed_names.add(n)
                added_names.add(n)

        if removed_names or added_names:
            log.msg("adding %d new slaves, removing %d" %
                    (len(added_names), len(removed_names)))

            for n in removed_names:
                slave = old_by_name[n]

                del self.slaves[n]
                slave.master = None
                slave.botmaster = None

                wfd = defer.waitForDeferred(
                    defer.maybeDeferred(lambda :
                        slave.disownServiceParent()))
                yield wfd
                wfd.getResult()

            for n in added_names:
                slave = new_by_name[n]
                slave.setServiceParent(self)

                slave.botmaster = self
                slave.master = self.master
                self.slaves[n] = slave

        metrics.MetricCountEvent.log("num_slaves",
                len(self.slaves), absolute=True)

        timer.stop()


    @defer.deferredGenerator
    def reconfigServiceBuilders(self, new_config):

        timer = metrics.Timer("BotMaster.reconfigServiceBuilders")
        timer.start()

        # arrange builders by name
        old_by_name = dict([ (b.name, b)
                            for b in list(self)
                            if isinstance(b, Builder) ])
        old_set = set(old_by_name.iterkeys())
        new_by_name = dict([ (bc.name, bc)
                            for bc in new_config.builders ])
        new_set = set(new_by_name.iterkeys())

        # calculate new builders, by name, and removed builders
        removed_names, added_names = util.diffSets(old_set, new_set)

        if removed_names or added_names:
            log.msg("adding %d new builders, removing %d" %
                    (len(added_names), len(removed_names)))

            for n in removed_names:
                builder = old_by_name[n]

                del self.builders[n]
                builder.master = None
                builder.botmaster = None

                wfd = defer.waitForDeferred(
                    defer.maybeDeferred(lambda :
                        builder.disownServiceParent()))
                yield wfd
                wfd.getResult()

            for n in added_names:
                builder = Builder(n)
                self.builders[n] = builder

                builder.botmaster = self
                builder.master = self.master
                builder.setServiceParent(self)

        self.builderNames = self.builders.keys()

        metrics.MetricCountEvent.log("num_builders",
                len(self.builders), absolute=True)

        timer.stop()


    def stopService(self):
        if self.buildrequest_sub:
            self.buildrequest_sub.unsubscribe()
            self.buildrequest_sub = None
        for b in self.builders.values():
            b.builder_status.addPointEvent(["master", "shutdown"])
            b.builder_status.saveYourself()
        return service.MultiService.stopService(self)

    def getLockByID(self, lockid):
        """Convert a Lock identifier into an actual Lock instance.
        @param lockid: a locks.MasterLock or locks.SlaveLock instance
        @return: a locks.RealMasterLock or locks.RealSlaveLock instance
        """
        assert isinstance(lockid, (locks.MasterLock, locks.SlaveLock))
        if not lockid in self.locks:
            self.locks[lockid] = lockid.lockClass(lockid)
        # if the master.cfg file has changed maxCount= on the lock, the next
        # time a build is started, they'll get a new RealLock instance. Note
        # that this requires that MasterLock and SlaveLock (marker) instances
        # be hashable and that they should compare properly.
        return self.locks[lockid]

    def maybeStartBuildsForBuilder(self, buildername):
        """
        Call this when something suggests that a particular builder may now
        be available to start a build.

        @param buildername: the name of the builder
        """
        self.brd.maybeStartBuildsOn([buildername])

    def maybeStartBuildsForSlave(self, slave_name):
        """
        Call this when something suggests that a particular slave may now be
        available to start a build.

        @param slave_name: the name of the slave
        """
        builders = self.getBuildersForSlave(slave_name)
        self.brd.maybeStartBuildsOn([ b.name for b in builders ])

    def maybeStartBuildsForAllBuilders(self):
        """
        Call this when something suggests that this would be a good time to start some
        builds, but nothing more specific.
        """
        self.brd.maybeStartBuildsOn(self.builderNames)

class BuildRequestDistributor(service.Service):
    """
    Special-purpose class to handle distributing build requests to builders by
    calling their C{maybeStartBuild} method.

    This takes account of the C{prioritizeBuilders} configuration, and is
    highly re-entrant; that is, if a new build request arrives while builders
    are still working on the previous build request, then this class will
    correctly re-prioritize invocations of builders' C{maybeStartBuild}
    methods.
    """

    def __init__(self, botmaster):
        self.botmaster = botmaster
        self.master = botmaster.master

        # lock to ensure builders are only sorted once at any time
        self.pending_builders_lock = defer.DeferredLock()

        # sorted list of names of builders that need their maybeStartBuild
        # method invoked.
        self._pending_builders = []
        self.activity_lock = defer.DeferredLock()
        self.active = False

    def stopService(self):
        # let the parent stopService succeed between activity; then the loop
        # will stop calling itself, since self.running is false
        d = self.activity_lock.acquire()
        d.addCallback(lambda _ : service.Service.stopService(self))
        d.addBoth(lambda _ : self.activity_lock.release())
        return d

    @defer.deferredGenerator
    def maybeStartBuildsOn(self, new_builders):
        """
        Try to start any builds that can be started right now.  This function
        returns immediately, and promises to trigger those builders
        eventually.

        @param new_builders: names of new builders that should be given the
        opportunity to check for new requests.
        """
        new_builders = set(new_builders)
        existing_pending = set(self._pending_builders)

        # if we won't add any builders, there's nothing to do
        if new_builders < existing_pending:
            return

        # reset the list of pending builders; this is async, so begin
        # by grabbing a lock
        wfd = defer.waitForDeferred(
            self.pending_builders_lock.acquire())
        yield wfd
        wfd.getResult()

        try:
            # re-fetch existing_pending, in case it has changed while acquiring
            # the lock
            existing_pending = set(self._pending_builders)

            # then sort the new, expanded set of builders
            wfd = defer.waitForDeferred(
                self._sortBuilders(list(existing_pending | new_builders)))
            yield wfd
            self._pending_builders = wfd.getResult()

            # start the activity loop, if we aren't already working on that.
            if not self.active:
                self._activityLoop()
        except:
            log.err(Failure(),
                    "while attempting to start builds on %s" % self.name)

        # release the lock unconditionally
        self.pending_builders_lock.release()

    @defer.deferredGenerator
    def _defaultSorter(self, master, builders):
        timer = metrics.Timer("BuildRequestDistributor._defaultSorter()")
        timer.start()
        # perform an asynchronous schwarzian transform, transforming None
        # into sys.maxint so that it sorts to the end
        def xform(bldr):
            d = defer.maybeDeferred(lambda :
                    bldr.getOldestRequestTime())
            d.addCallback(lambda time :
                (((time is None) and None or time),bldr))
            return d
        wfd = defer.waitForDeferred(
            defer.gatherResults(
                [ xform(bldr) for bldr in builders ]))
        yield wfd
        xformed = wfd.getResult()

        # sort the transformed list synchronously, comparing None to the end of
        # the list
        def nonecmp(a,b):
            if a[0] is None: return 1
            if b[0] is None: return -1
            return cmp(a,b)
        xformed.sort(cmp=nonecmp)

        # and reverse the transform
        yield [ xf[1] for xf in xformed ]
        timer.stop()

    @defer.deferredGenerator
    def _sortBuilders(self, buildernames):
        timer = metrics.Timer("BuildRequestDistributor._sortBuilders()")
        timer.start()
        # note that this takes and returns a list of builder names

        # convert builder names to builders
        builders_dict = self.botmaster.builders
        builders = [ builders_dict.get(n)
                     for n in buildernames
                     if n in builders_dict ]

        # find a sorting function
        sorter = self.master.config.prioritizeBuilders
        if not sorter:
            sorter = self._defaultSorter

        # run it
        try:
            wfd = defer.waitForDeferred(
                defer.maybeDeferred(lambda :
                    sorter(self.master, builders)))
            yield wfd
            builders = wfd.getResult()
        except:
            log.msg("Exception prioritizing builders; order unspecified")
            log.err(Failure())

        # and return the names
        yield [ b.name for b in builders ]
        timer.stop()

    @defer.deferredGenerator
    def _activityLoop(self):
        self.active = True

        timer = metrics.Timer('BuildRequestDistributor._activityLoop()')
        timer.start()

        while 1:
            wfd = defer.waitForDeferred(
                self.activity_lock.acquire())
            yield wfd
            wfd.getResult()

            # lock pending_builders, pop an element from it, and release
            wfd = defer.waitForDeferred(
                self.pending_builders_lock.acquire())
            yield wfd
            wfd.getResult()

            # bail out if we shouldn't keep looping
            if not self.running or not self._pending_builders:
                self.pending_builders_lock.release()
                self.activity_lock.release()
                break

            bldr_name = self._pending_builders.pop(0)
            self.pending_builders_lock.release()

            try:
                wfd = defer.waitForDeferred(
                    self._callABuilder(bldr_name))
                yield wfd
                wfd.getResult()
            except:
                log.err(Failure(),
                        "from maybeStartBuild for builder '%s'" % (bldr_name,))

            self.activity_lock.release()

        timer.stop()

        self.active = False
        self._quiet()

    def _callABuilder(self, bldr_name):
        # get the actual builder object
        bldr = self.botmaster.builders.get(bldr_name)
        if not bldr:
            return defer.succeed(None)

        d = bldr.maybeStartBuild()
        d.addErrback(log.err, 'in maybeStartBuild for %r' % (bldr,))
        return d

    def _quiet(self):
        # shim for tests
        pass # pragma: no cover


class DuplicateSlaveArbitrator(object):
    """Utility class to arbitrate the situation when a new slave connects with
    the name of an existing, connected slave

    @ivar buildslave: L{buildbot.process.slavebuilder.AbstractBuildSlave}
    instance
    @ivar old_remote: L{RemoteReference} to the old slave
    @ivar new_remote: L{RemoteReference} to the new slave
    """
    _reactor = reactor # for testing

    # There are several likely duplicate slave scenarios in practice:
    #
    # 1. two slaves are configured with the same username/password
    #
    # 2. the same slave process believes it is disconnected (due to a network
    # hiccup), and is trying to reconnect
    #
    # For the first case, we want to prevent the two slaves from repeatedly
    # superseding one another (which results in lots of failed builds), so we
    # will prefer the old slave.  However, for the second case we need to
    # detect situations where the old slave is "gone".  Sometimes "gone" means
    # that the TCP/IP connection to it is in a long timeout period (10-20m,
    # depending on the OS configuration), so this can take a while.

    PING_TIMEOUT = 10
    """Timeout for pinging the old slave.  Set this to something quite long, as
    a very busy slave (e.g., one sending a big log chunk) may take a while to
    return a ping.
    """

    def __init__(self, buildslave):
        self.buildslave = buildslave
        self.old_remote = self.buildslave.slave

    def getPerspective(self, mind, slavename):
        self.new_remote = mind
        self.ping_old_slave_done = False
        self.old_slave_connected = True
        self.ping_new_slave_done = False

        old_tport = self.old_remote.broker.transport
        new_tport = self.new_remote.broker.transport
        log.msg("duplicate slave %s; delaying new slave (%s) and pinging old "
                "(%s)" % (self.buildslave.slavename, new_tport.getPeer(),
                          old_tport.getPeer()))

        # delay the new slave until we decide what to do with it
        d = self.new_slave_d = defer.Deferred()

        # Ping the old slave.  If this kills it, then we can allow the new
        # slave to connect.  If this does not kill it, then we disconnect
        # the new slave.
        self.ping_old_slave(new_tport.getPeer())

        # Print a message on the new slave, if possible.
        self.ping_new_slave()

        return d

    def ping_new_slave(self):
        d = defer.maybeDeferred(lambda :
            self.new_remote.callRemote("print", "master already has a "
                        "connection named '%s' - checking its liveness"
                        % self.buildslave.slavename))
        def done(_):
            # failure or success, doesn't matter - the ping is done.
            self.ping_new_slave_done = True
            self.maybe_done()
        d.addBoth(done)

    def ping_old_slave(self, new_peer):
        # set a timer on this ping, in case the network is bad.  TODO: a
        # timeout on the ping itself is not quite what we want.  If there is
        # other data flowing over the PB connection, then we should keep
        # waiting.  Bug #1703
        def timeout():
            self.ping_old_slave_timeout = None
            self.ping_old_slave_timed_out = True
            self.old_slave_connected = False
            self.ping_old_slave_done = True
            self.maybe_done()
        self.ping_old_slave_timeout = self._reactor.callLater(
                                    self.PING_TIMEOUT, timeout)
        self.ping_old_slave_timed_out = False

        # call this in maybeDeferred because callRemote tends to raise
        # exceptions instead of returning Failures
        d = defer.maybeDeferred(lambda :
            self.old_remote.callRemote("print",
                "master got a duplicate connection from %s; keeping this one"
                                        % new_peer))

        def clear_timeout(r):
            if self.ping_old_slave_timeout:
                self.ping_old_slave_timeout.cancel()
                self.ping_old_slave_timeout = None
            return r
        d.addBoth(clear_timeout)

        def old_gone(f):
            if self.ping_old_slave_timed_out:
                return # ignore after timeout
            f.trap(pb.PBConnectionLost, pb.DeadReferenceError)
            log.msg(("connection lost while pinging old slave '%s' - " +
                     "keeping new slave") % self.buildslave.slavename)
            self.old_slave_connected = False
        d.addErrback(old_gone)

        def other_err(f):
            log.err(f, "unexpected error pinging old slave; disconnecting it")
            self.old_slave_connected = False
        d.addErrback(other_err)

        def done(_):
            if self.ping_old_slave_timed_out:
                return # ignore after timeout
            self.ping_old_slave_done = True
            self.maybe_done()
        d.addCallback(done)

    def maybe_done(self):
        if not self.ping_new_slave_done or not self.ping_old_slave_done:
            return

        # both pings are done, so sort out the results
        if self.old_slave_connected:
            self.disconnect_new_slave()
        else:
            self.start_new_slave()

    def start_new_slave(self):
        # just in case
        if not self.new_slave_d: # pragma: ignore
            return

        d = self.new_slave_d
        self.new_slave_d = None

        if self.buildslave.isConnected():
            # we need to wait until the old slave has fully detached, which can
            # take a little while as buffers drain, etc.
            def detached():
                d.callback(self.buildslave)
            self.buildslave.subscribeToDetach(detached)
            self.old_remote.broker.transport.loseConnection()
        else: # pragma: ignore
            # by some unusual timing, it's quite possible that the old slave
            # has disconnected while the arbitration was going on.  In that
            # case, we're already done!
            d.callback(self.buildslave)

    def disconnect_new_slave(self):
        # just in case
        if not self.new_slave_d: # pragma: ignore
            return

        d = self.new_slave_d
        self.new_slave_d = None
        log.msg("rejecting duplicate slave with exception")
        d.errback(Failure(RuntimeError("rejecting duplicate slave")))





# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

SUCCESS, WARNINGS, FAILURE, SKIPPED, EXCEPTION, RETRY = range(6)
Results = ["success", "warnings", "failure", "skipped", "exception", "retry"]

def worst_status(a, b):
    # SUCCESS > WARNINGS > FAILURE > EXCEPTION > RETRY
    # Retry needs to be considered the worst so that conusmers don't have to
    # worry about other failures undermining the RETRY.
    for s in (RETRY, EXCEPTION, FAILURE, WARNINGS, SKIPPED, SUCCESS):
        if s in (a, b):
            return s



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.python import failure
from twisted.internet import defer
from buildbot.schedulers import base
from buildbot.process.properties import Properties

class Triggerable(base.BaseScheduler):

    compare_attrs = base.BaseScheduler.compare_attrs

    def __init__(self, name, builderNames, properties={}):
        base.BaseScheduler.__init__(self, name, builderNames, properties)
        self._waiters = {}
        self._bsc_subscription = None
        self.reason = "Triggerable(%s)" % name

    def trigger(self, ss_setid, set_props=None):
        """Trigger this scheduler with the given sourcestampset ID. Returns a
        deferred that will fire when the buildset is finished."""
        # properties for this buildset are composed of our own properties,
        # potentially overridden by anything from the triggering build
        props = Properties()
        props.updateFromProperties(self.properties)
        if set_props:
            props.updateFromProperties(set_props)

        # note that this does not use the buildset subscriptions mechanism, as
        # the duration of interest to the caller is bounded by the lifetime of
        # this process.
        if ss_setid:
            d = self.addBuildsetForSourceStamp(reason=self.reason, setid=ss_setid,
                    properties=props)
        else:
            d = self.addBuildsetForLatest(reason=self.reason, properties=props)
        def setup_waiter((bsid,brids)):
            d = defer.Deferred()
            self._waiters[bsid] = (d, brids)
            self._updateWaiters()
            return d
        d.addCallback(setup_waiter)
        return d

    def stopService(self):
        # cancel any outstanding subscription
        if self._bsc_subscription:
            self._bsc_subscription.unsubscribe()
            self._bsc_subscription = None

        # and errback any outstanding deferreds
        if self._waiters:
            msg = 'Triggerable scheduler stopped before build was complete'
            for d, brids in self._waiters.values():
                d.errback(failure.Failure(RuntimeError(msg)))
            self._waiters = {}

        return base.BaseScheduler.stopService(self)

    def _updateWaiters(self):
        if self._waiters and not self._bsc_subscription:
            self._bsc_subscription = \
                    self.master.subscribeToBuildsetCompletions(
                                                self._buildsetComplete)
        elif not self._waiters and self._bsc_subscription:
            self._bsc_subscription.unsubscribe()
            self._bsc_subscription = None

    def _buildsetComplete(self, bsid, result):
        if bsid not in self._waiters:
            return

        # pop this bsid from the waiters list, and potentially unsubscribe
        # from completion notifications
        d, brids = self._waiters.pop(bsid)
        self._updateWaiters()

        # fire the callback to indicate that the triggered build is complete
        d.callback((result, brids))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import traceback
import re
from twisted.internet import defer
try:
    import email.utils as email_utils
    email_utils = email_utils
except ImportError:
    # Python-2.4 capitalization
    import email.Utils as email_utils

from buildbot.process.properties import Properties
from buildbot.schedulers import base

class BaseParameter(object):
    name = ""
    label = ""
    type = ""
    default = ""
    required = False
    multiple = False
    regex = None
    debug=True

    def __init__(self, name, label=None, regex=None, **kw):
        self.label = self.name = name
        if label:
            self.label = label
        if regex:
            self.regex = re.compile(regex)
        # all other properties are generically passed via **kw
        self.__dict__.update(kw)

    def update_from_post(self, master, properties, changes, req):
        args = req.args.get(self.name, [])
        if len(args) == 0:
            if self.required:
                raise ValueError("'%s' needs to be specified" % (self.label))
            if self.multiple:
                args = self.default
            else:
                args = [self.default]
        if self.regex:
            for arg in args:
                if not self.regex.match(arg):
                    raise ValueError("%s:'%s' does not match pattern '%s'"
                            % (self.label, arg, self.regex.pattern))
        try:
            arg = self.parse_from_args(args)
        except Exception, e:
            # an exception will just display an alert in the web UI
            # also log the exception
            if self.debug:
                traceback.print_exc()
            raise e
        if arg == None:
            raise ValueError("need %s: no default provided by config"
                    % (self.name,))
        properties[self.name] = arg

    def parse_from_args(self, l):
        if self.multiple:
            return map(self.parse_from_arg, l)
        else:
            return self.parse_from_arg(l[0])

    def parse_from_arg(self, s):
        return s


class FixedParameter(BaseParameter):
    type = "fixed"
    hide = True
    default = ""

    def parse_from_args(self, l):
        return self.default


class StringParameter(BaseParameter):
    type = "text"
    size = 10

    def parse_from_arg(self, s):
        return s


class TextParameter(StringParameter):
    type = "textarea"
    cols = 80
    rows = 20

    def value_to_text(self, value):
        return str(value)


class IntParameter(StringParameter):
    type = "int"

    parse_from_arg = int # will throw an exception if parse fail


class BooleanParameter(BaseParameter):
    type = "bool"

    def update_from_post(self, master, properties, changes, req):
        # damn html's ungeneric checkbox implementation...
        checkbox = req.args.get("checkbox", [""])
        properties[self.name] = self.name in checkbox


class UserNameParameter(StringParameter):
    type = "text"
    default = ""
    size = 30
    need_email = True

    def __init__(self, name="username", label="Your name:", **kw):
        BaseParameter.__init__(self, name, label, **kw)

    def parse_from_arg(self, s):
        if self.need_email:
            e = email_utils.parseaddr(s)
            if e[0]=='' or e[1] == '':
                raise ValueError("%s: please fill in email address in the "
                        " form User <email@email.com>" % (self.label,))
        return s


class ChoiceStringParameter(BaseParameter):
    type = "list"
    choices = []
    strict = True

    def parse_from_arg(self, s):
        if self.strict and not s in self.choices:
            raise ValueError("'%s' does not belongs to list of available choices '%s'"%(s, self.choices))
        return s


class InheritBuildParameter(ChoiceStringParameter):
    name = "inherit"
    compatible_builds = None

    def update_from_post(self, master, properties, changes, req):
        arg = req.args.get(self.name, [""])[0]
        splitted_arg = arg.split(" ")[0].split("/")
        if len(splitted_arg) != 2:
            raise ValueError("bad build: %s"%(arg))
        builder, num = splitted_arg
        builder_status = master.status.getBuilder(builder)
        if not builder_status:
            raise ValueError("unknown builder: %s in %s"%(builder, arg))
        b = builder_status.getBuild(int(num))
        if not b:
            raise ValueError("unknown build: %d in %s"%(num, arg))
        props = {self.name:(arg.split(" ")[0])}
        for name, value, source in b.getProperties().asList():
            if source == "Force Build Form":
                if name == "owner":
                    name = "orig_owner"
                props[name] = value
        properties.update(props)
        changes.extend(b.changes)


class AnyPropertyParameter(BaseParameter):
    type = "anyproperty"

    def update_from_post(self, master, properties, changes, req):
        validation = master.config.validation
        pname = req.args.get("%sname" % self.name, [""])[0]
        pvalue = req.args.get("%svalue" % self.name, [""])[0]
        if not pname:
            return
        pname_validate = validation['property_name']
        pval_validate = validation['property_value']
        if not pname_validate.match(pname) \
                or not pval_validate.match(pvalue):
            raise ValueError("bad property name='%s', value='%s'" % (pname, pvalue))
        properties[pname] = pvalue


class ForceScheduler(base.BaseScheduler):
    
    compare_attrs = ( 'name', 'builderNames', 'branch', 'reason',
            'revision', 'repository', 'project', 'forcedProperties' )

    def __init__(self, name, builderNames,
            branch=StringParameter(name="branch",default=""), 
            reason=StringParameter(name="reason", default="force build"),
            revision=StringParameter(name="revision",default=""),
            repository=StringParameter(name="repository",default=""),
            project=StringParameter(name="project",default=""),
            username=UserNameParameter(),
            properties=[
                AnyPropertyParameter("property1"),
                AnyPropertyParameter("property2"),
                AnyPropertyParameter("property3"),
                AnyPropertyParameter("property4"),
            ]):

        base.BaseScheduler.__init__(self, name=name,
                builderNames=builderNames,properties={})
        self.branch = branch
        self.reason = reason
        self.repository = repository
        self.revision = revision
        self.project = project
        self.username = username
        self.forcedProperties = properties
        # this is used to simplify the template
        self.all_fields = [ branch, username, reason, repository,
                            revision, project ]
        self.all_fields.extend(properties)

    def startService(self):
        pass

    def stopService(self):
        pass

    def forceWithWebRequest(self, owner, builder_name, req):
        """Called by the web UI.
        Authentication is already done, thus owner is passed as argument
        We check the parameters, and launch the build, if everything is correct
        """
        if not builder_name in self.builderNames:
            # in the case of buildAll, this method will be called several times
            # for all the builders
            # we just do nothing on a builder that is not in our builderNames
            return defer.succeed(None)
        master = self.master
        properties = {}
        changeids = []
        # probably need to clean that out later as the IProperty is already a
        # validation mechanism

        validation = master.config.validation
        if self.branch.regex == None:
            self.branch.regex = validation['branch']
        if self.revision.regex == None:
            self.revision.regex = validation['revision']

        for param in self.all_fields:
            if owner and param==self.username:
                continue # dont enforce username if auth
            param.update_from_post(master, properties, changeids, req)

        changeids = map(lambda a: type(a)==int and a or a.number, changeids)
        # everything is validated, we can create our source stamp, and buildrequest
        reason = properties[self.reason.name]
        branch = properties[self.branch.name]
        revision = properties[self.revision.name]
        repository = properties[self.repository.name]
        project = properties[self.project.name]
        if owner is None:
            owner =  properties[self.username.name]

        std_prop_names = [self.branch.name, 
                          self.revision.name, self.repository.name,
                          self.project.name, self.username.name]
        real_properties = Properties()
        for pname, pvalue in properties.items():
            if not pname in std_prop_names:
                real_properties.setProperty(pname, pvalue, "Force Build Form")

        real_properties.setProperty("owner", owner, "Force Build Form")

        r = ("The web-page 'force build' button was pressed by '%s': %s\n"
             % (owner, reason)) 

        d = master.db.sourcestampsets.addSourceStampSet()
        def add_master_with_setid(sourcestampsetid):
            master.db.sourcestamps.addSourceStamp(
                                    sourcestampsetid = sourcestampsetid,
                                    branch=branch,
                                    revision=revision, project=project, 
                                    repository=repository,changeids=changeids)
            return sourcestampsetid
            
        d.addCallback(add_master_with_setid)
        def got_setid(sourcestampsetid):
            return self.addBuildsetForSourceStamp(builderNames=[builder_name],
                                    setid=sourcestampsetid, reason=r,
                                    properties=real_properties)
        d.addCallback(got_setid)
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re
import os
import textwrap
import mock
from zope.interface import implements
from twisted.trial import unittest
from twisted.application import service
from twisted.internet import defer
from buildbot import config, buildslave, interfaces, revlinks
from buildbot.process import properties
from buildbot.test.util import dirs, compat
from buildbot.changes import base as changes_base
from buildbot.schedulers import base as schedulers_base
from buildbot.status import base as status_base

global_defaults = dict(
    title='Buildbot',
    titleURL='http://buildbot.net',
    buildbotURL='http://localhost:8080/',
    changeHorizon=None,
    eventHorizon=50,
    logHorizon=None,
    buildHorizon=None,
    logCompressionLimit=4096,
    logCompressionMethod='bz2',
    logMaxTailSize=None,
    logMaxSize=None,
    properties=properties.Properties(),
    mergeRequests=None,
    prioritizeBuilders=None,
    slavePortnum=None,
    multiMaster=False,
    debugPassword=None,
    manhole=None,
)


class FakeChangeSource(changes_base.ChangeSource):
    pass


class FakeStatusReceiver(status_base.StatusReceiver):
    pass


class FakeScheduler(object):
    implements(interfaces.IScheduler)
    def __init__(self, name):
        self.name = name

class FakeBuilder(object):

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class ConfigErrorsMixin(object):

    def assertConfigError(self, errors, substr_or_re):
        if len(errors.errors) > 1:
            self.fail("too many errors: %s" % (errors.errors,))
        elif len(errors.errors) < 1:
            self.fail("expected error did not occur")
        elif isinstance(substr_or_re, str):
            if substr_or_re not in errors.errors[0]:
                self.fail("non-matching error: %s" % (errors.errors,))
        else:
            if not substr_or_re.search(errors.errors[0]):
                self.fail("non-matching error: %s" % (errors.errors,))

    def assertRaisesConfigError(self, substr_or_re, fn):
        try:
            fn()
        except config.ConfigErrors, e:
            self.assertConfigError(e, substr_or_re)
        else:
            self.fail("ConfigErrors not raised")

    def assertNoConfigErrors(self, errors):
        self.assertEqual(errors.errors, [])


class ConfigErrors(unittest.TestCase):

    def test_constr(self):
        ex = config.ConfigErrors(['a', 'b'])
        self.assertEqual(ex.errors, ['a', 'b'])

    def test_addError(self):
        ex = config.ConfigErrors(['a'])
        ex.addError('c')
        self.assertEqual(ex.errors, ['a', 'c'])

    def test_nonempty(self):
        empty = config.ConfigErrors()
        full = config.ConfigErrors(['a'])
        self.failUnless(not empty)
        self.failIf(not full)

    def test_error_raises(self):
        e = self.assertRaises(config.ConfigErrors, config.error, "message")
        self.assertEqual(e.errors, ["message"])

    def test_error_no_raise(self):
        e = config.ConfigErrors()
        self.patch(config, "_errors", e)
        config.error("message")
        self.assertEqual(e.errors, ["message"])


class MasterConfig(ConfigErrorsMixin, dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        self.basedir = os.path.abspath('basedir')
        self.filename = 'test.cfg'
        return self.setUpDirs('basedir')

    def tearDown(self):
        return self.tearDownDirs()

    # utils

    def patch_load_helpers(self):
        # patch out all of the "helpers" for laodConfig with null functions
        for n in dir(config.MasterConfig):
            if n.startswith('load_'):
                typ = 'loader'
            elif n.startswith('check_'):
                typ = 'checker'
            else:
                continue

            v = getattr(config.MasterConfig, n)
            if callable(v):
                if typ == 'loader':
                    self.patch(config.MasterConfig, n,
                        mock.Mock(side_effect=
                            lambda filename, config_dict, errors : None))
                else:
                    self.patch(config.MasterConfig, n,
                        mock.Mock(side_effect=
                            lambda errors : None))


    def install_config_file(self, config_file, other_files={}):
        config_file = textwrap.dedent(config_file)
        open(os.path.join(self.basedir, self.filename), "w").write(config_file)
        for file, contents in other_files.items():
            open(file, "w").write(contents)


    # tests

    def test_defaults(self):
        cfg = config.MasterConfig()
        expected = dict(
            #validation,
            db=dict(
                db_url='sqlite:///state.sqlite',
                db_poll_interval=None),
            metrics = None,
            caches = dict(Changes=10, Builds=15),
            schedulers = {},
            builders = [],
            slaves = [],
            change_sources = [],
            status = [],
            user_managers = [],
            revlink = revlinks.default_revlink_matcher
            )
        expected.update(global_defaults)
        got = dict([
            (attr, getattr(cfg, attr))
            for attr, exp in expected.iteritems() ])
        self.assertEqual(got, expected)

    def test_defaults_validation(self):
        # re's aren't comparable, but we can make sure the keys match
        cfg = config.MasterConfig()
        self.assertEqual(sorted(cfg.validation.keys()),
            sorted([
                'branch', 'revision', 'property_name', 'property_value',
            ]))

    def test_loadConfig_missing_file(self):
        self.assertRaisesConfigError(
            re.compile("configuration file .* does not exist"),
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))

    def test_loadConfig_missing_basedir(self):
        self.assertRaisesConfigError(
            re.compile("basedir .* does not exist"),
            lambda : config.MasterConfig.loadConfig(
                os.path.join(self.basedir, 'NO'), 'test.cfg'))

    @compat.usesFlushLoggedErrors
    def test_loadConfig_parse_error(self):
        self.install_config_file('def x:\nbar')
        self.assertRaisesConfigError(
            re.compile("error while parsing.*traceback in logfile"),
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))
        self.assertEqual(len(self.flushLoggedErrors(SyntaxError)), 1)

    def test_loadConfig_eval_ConfigError(self):
        self.install_config_file("""\
                from buildbot import config
                BuildmasterConfig = { 'multiMaster': True }
                config.error('oh noes!')""")
        self.assertRaisesConfigError("oh noes",
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))

    def test_loadConfig_eval_ConfigErrors(self):
        # We test a config that has embedded errors, as well
        # as semantic errors that get added later. If an exception is raised
        # prematurely, then the semantic errors wouldn't get reported.
        self.install_config_file("""\
                from buildbot import config
                BuildmasterConfig = {}
                config.error('oh noes!')
                config.error('noes too!')""")
        e = self.assertRaises(config.ConfigErrors,
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))
        self.assertEqual(e.errors, ['oh noes!', 'noes too!',
                'no slaves are configured',
                'no builders are configured'])

    def test_loadConfig_no_BuildmasterConfig(self):
        self.install_config_file('x=10')
        self.assertRaisesConfigError("does not define 'BuildmasterConfig'",
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))

    def test_loadConfig_unknown_key(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict(foo=10)
                """)
        self.assertRaisesConfigError("Unknown BuildmasterConfig key foo",
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))

    def test_loadConfig_unknown_keys(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict(foo=10, bar=20)
                """)
        self.assertRaisesConfigError("Unknown BuildmasterConfig keys bar, foo",
            lambda : config.MasterConfig.loadConfig(
                self.basedir, self.filename))

    def test_loadConfig_success(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict()
                """)
        rv = config.MasterConfig.loadConfig(
            self.basedir, self.filename)
        self.assertIsInstance(rv, config.MasterConfig)

        # make sure all of the loaders and checkers are called
        self.failUnless(rv.load_global.called)
        self.failUnless(rv.load_validation.called)
        self.failUnless(rv.load_db.called)
        self.failUnless(rv.load_metrics.called)
        self.failUnless(rv.load_caches.called)
        self.failUnless(rv.load_schedulers.called)
        self.failUnless(rv.load_builders.called)
        self.failUnless(rv.load_slaves.called)
        self.failUnless(rv.load_change_sources.called)
        self.failUnless(rv.load_status.called)
        self.failUnless(rv.load_user_managers.called)

        self.failUnless(rv.check_single_master.called)
        self.failUnless(rv.check_schedulers.called)
        self.failUnless(rv.check_locks.called)
        self.failUnless(rv.check_builders.called)
        self.failUnless(rv.check_status.called)
        self.failUnless(rv.check_horizons.called)
        self.failUnless(rv.check_slavePortnum.called)

    def test_loadConfig_with_local_import(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                from subsidiary_module import x
                BuildmasterConfig = dict()
                """,
                {'basedir/subsidiary_module.py' : "x = 10"})
        rv = config.MasterConfig.loadConfig(
            self.basedir, self.filename)
        self.assertIsInstance(rv, config.MasterConfig)


class MasterConfig_loaders(ConfigErrorsMixin, unittest.TestCase):

    filename = 'test.cfg'

    def setUp(self):
        self.cfg = config.MasterConfig()
        self.errors = config.ConfigErrors()

    # utils

    def assertResults(self, **expected):
        self.failIf(self.errors, self.errors.errors)
        got = dict([
            (attr, getattr(self.cfg, attr))
            for attr, exp in expected.iteritems() ])
        self.assertEqual(got, expected)

    # tests

    def test_load_global_defaults(self):
        self.cfg.load_global(self.filename, {}, self.errors)
        self.assertResults(**global_defaults)

    def test_load_global_string_param_not_string(self):
        self.cfg.load_global(self.filename,
                dict(title=10), self.errors)
        self.assertConfigError(self.errors, 'must be a string')

    def test_load_global_int_param_not_int(self):
        self.cfg.load_global(self.filename,
                dict(changeHorizon='yes'), self.errors)
        self.assertConfigError(self.errors, 'must be an int')

    def do_test_load_global(self, config_dict, **expected):
        self.cfg.load_global(self.filename, config_dict, self.errors)
        self.assertResults(**expected)

    def test_load_global_title(self):
        self.do_test_load_global(dict(title='hi'), title='hi')

    def test_load_global_projectURL(self):
        self.do_test_load_global(dict(projectName='hey'), title='hey')

    def test_load_global_titleURL(self):
        self.do_test_load_global(dict(titleURL='hi'), titleURL='hi')

    def test_load_global_buildbotURL(self):
        self.do_test_load_global(dict(buildbotURL='hey'), buildbotURL='hey')

    def test_load_global_changeHorizon(self):
        self.do_test_load_global(dict(changeHorizon=10), changeHorizon=10)

    def test_load_global_changeHorizon_none(self):
        self.do_test_load_global(dict(changeHorizon=None), changeHorizon=None)

    def test_load_global_eventHorizon(self):
        self.do_test_load_global(dict(eventHorizon=10), eventHorizon=10)

    def test_load_global_logHorizon(self):
        self.do_test_load_global(dict(logHorizon=10), logHorizon=10)

    def test_load_global_buildHorizon(self):
        self.do_test_load_global(dict(buildHorizon=10), buildHorizon=10)

    def test_load_global_logCompressionLimit(self):
        self.do_test_load_global(dict(logCompressionLimit=10),
                                 logCompressionLimit=10)

    def test_load_global_logCompressionMethod(self):
        self.do_test_load_global(dict(logCompressionMethod='gz'),
                                 logCompressionMethod='gz')

    def test_load_global_logCompressionMethod_invalid(self):
        self.cfg.load_global(self.filename,
                dict(logCompressionMethod='foo'), self.errors)
        self.assertConfigError(self.errors, "must be 'bz2' or 'gz'")

    def test_load_global_logMaxSize(self):
        self.do_test_load_global(dict(logMaxSize=123), logMaxSize=123)

    def test_load_global_logMaxTailSize(self):
        self.do_test_load_global(dict(logMaxTailSize=123), logMaxTailSize=123)

    def test_load_global_properties(self):
        exp = properties.Properties()
        exp.setProperty('x', 10, self.filename)
        self.do_test_load_global(dict(properties=dict(x=10)), properties=exp)

    def test_load_global_properties_invalid(self):
        self.cfg.load_global(self.filename,
                dict(properties='yes'), self.errors)
        self.assertConfigError(self.errors, "must be a dictionary")

    def test_load_global_mergeRequests_bool(self):
        self.do_test_load_global(dict(mergeRequests=False),
                mergeRequests=False)

    def test_load_global_mergeRequests_callable(self):
        callable = lambda : None
        self.do_test_load_global(dict(mergeRequests=callable),
                mergeRequests=callable)

    def test_load_global_mergeRequests_invalid(self):
        self.cfg.load_global(self.filename,
                dict(mergeRequests='yes'), self.errors)
        self.assertConfigError(self.errors,
                "must be a callable, True, or False")

    def test_load_global_prioritizeBuilders_callable(self):
        callable = lambda : None
        self.do_test_load_global(dict(prioritizeBuilders=callable),
                prioritizeBuilders=callable)

    def test_load_global_prioritizeBuilders_invalid(self):
        self.cfg.load_global(self.filename,
                dict(prioritizeBuilders='yes'), self.errors)
        self.assertConfigError(self.errors, "must be a callable")

    def test_load_global_slavePortnum_int(self):
        self.do_test_load_global(dict(slavePortnum=123),
                slavePortnum='tcp:123')

    def test_load_global_slavePortnum_str(self):
        self.do_test_load_global(dict(slavePortnum='udp:123'),
                slavePortnum='udp:123')

    def test_load_global_multiMaster(self):
        self.do_test_load_global(dict(multiMaster=1), multiMaster=1)

    def test_load_global_debugPassword(self):
        self.do_test_load_global(dict(debugPassword='xyz'),
                debugPassword='xyz')

    def test_load_global_manhole(self):
        mh = mock.Mock(name='manhole')
        self.do_test_load_global(dict(manhole=mh), manhole=mh)

    def test_load_global_revlink_callable(self):
        callable = lambda : None
        self.do_test_load_global(dict(revlink=callable),
                revlink=callable)

    def test_load_global_revlink_invalid(self):
        self.cfg.load_global(self.filename, dict(revlink=''), self.errors)
        self.assertConfigError(self.errors, "must be a callable")

    def test_load_validation_defaults(self):
        self.cfg.load_validation(self.filename, {}, self.errors)
        self.assertEqual(sorted(self.cfg.validation.keys()),
            sorted([
                'branch', 'revision', 'property_name', 'property_value',
            ]))

    def test_load_validation_invalid(self):
        self.cfg.load_validation(self.filename,
                dict(validation='plz'), self.errors)
        self.assertConfigError(self.errors, "must be a dictionary")

    def test_load_validation_unk_keys(self):
        self.cfg.load_validation(self.filename,
                dict(validation=dict(users='.*')), self.errors)
        self.assertConfigError(self.errors, "unrecognized validation key(s)")

    def test_load_validation(self):
        r = re.compile('.*')
        self.cfg.load_validation(self.filename,
                dict(validation=dict(branch=r)), self.errors)
        self.assertEqual(self.cfg.validation['branch'], r)
        # check that defaults are still around
        self.assertIn('revision', self.cfg.validation)


    def test_load_db_defaults(self):
        self.cfg.load_db(self.filename, {}, self.errors)
        self.assertResults(
            db=dict(db_url='sqlite:///state.sqlite', db_poll_interval=None))

    def test_load_db_db_url(self):
        self.cfg.load_db(self.filename, dict(db_url='abcd'), self.errors)
        self.assertResults(db=dict(db_url='abcd', db_poll_interval=None))

    def test_load_db_db_poll_interval(self):
        self.cfg.load_db(self.filename, dict(db_poll_interval=2), self.errors)
        self.assertResults(
            db=dict(db_url='sqlite:///state.sqlite', db_poll_interval=2))

    def test_load_db_dict(self):
        self.cfg.load_db(self.filename,
            dict(db=dict(db_url='abcd', db_poll_interval=10)),
            self.errors)
        self.assertResults(db=dict(db_url='abcd', db_poll_interval=10))

    def test_load_db_unk_keys(self):
        self.cfg.load_db(self.filename,
            dict(db=dict(db_url='abcd', db_poll_interval=10, bar='bar')),
            self.errors)
        self.assertConfigError(self.errors, "unrecognized keys in")

    def test_load_db_not_int(self):
        self.cfg.load_db(self.filename,
            dict(db=dict(db_url='abcd', db_poll_interval='ten')),
            self.errors)
        self.assertConfigError(self.errors, "must be an int")


    def test_load_metrics_defaults(self):
        self.cfg.load_metrics(self.filename, {}, self.errors)
        self.assertResults(metrics=None)

    def test_load_metrics_invalid(self):
        self.cfg.load_metrics(self.filename, dict(metrics=13), self.errors)
        self.assertConfigError(self.errors, "must be a dictionary")

    def test_load_metrics(self):
        self.cfg.load_metrics(self.filename,
                dict(metrics=dict(foo=1)),
                self.errors)
        self.assertResults(metrics=dict(foo=1))


    def test_load_caches_defaults(self):
        self.cfg.load_caches(self.filename, {}, self.errors)
        self.assertResults(caches=dict(Changes=10, Builds=15))

    def test_load_caches_invalid(self):
        self.cfg.load_caches(self.filename, dict(caches=13), self.errors)
        self.assertConfigError(self.errors, "must be a dictionary")

    def test_load_caches_buildCacheSize(self):
        self.cfg.load_caches(self.filename,
                dict(buildCacheSize=13),
                self.errors)
        self.assertResults(caches=dict(Builds=13, Changes=10))

    def test_load_caches_buildCacheSize_and_caches(self):
        self.cfg.load_caches(self.filename,
                dict(buildCacheSize=13, caches=dict(builds=11)),
                self.errors)
        self.assertConfigError(self.errors, "cannot specify")

    def test_load_caches_changeCacheSize(self):
        self.cfg.load_caches(self.filename,
                dict(changeCacheSize=13),
                self.errors)
        self.assertResults(caches=dict(Changes=13, Builds=15))

    def test_load_caches_changeCacheSize_and_caches(self):
        self.cfg.load_caches(self.filename,
                dict(changeCacheSize=13, caches=dict(changes=11)),
                self.errors)
        self.assertConfigError(self.errors, "cannot specify")

    def test_load_caches(self):
        self.cfg.load_caches(self.filename,
                dict(caches=dict(foo=1)),
                self.errors)
        self.assertResults(caches=dict(Changes=10, Builds=15, foo=1))


    def test_load_schedulers_defaults(self):
        self.cfg.load_schedulers(self.filename, {}, self.errors)
        self.assertResults(schedulers={})

    def test_load_schedulers_not_list(self):
        self.cfg.load_schedulers(self.filename,
                dict(schedulers=dict()), self.errors)
        self.assertConfigError(self.errors, "must be a list of")

    def test_load_schedulers_not_instance(self):
        self.cfg.load_schedulers(self.filename,
                dict(schedulers=[mock.Mock()]), self.errors)
        self.assertConfigError(self.errors, "must be a list of")

    def test_load_schedulers_dupe(self):
        sch1 = FakeScheduler(name='sch')
        sch2 = FakeScheduler(name='sch')
        self.cfg.load_schedulers(self.filename,
                dict(schedulers=[ sch1, sch2 ]), self.errors)
        self.assertConfigError(self.errors,
                "scheduler name 'sch' used multiple times")

    def test_load_schedulers(self):
        class Sch(schedulers_base.BaseScheduler):
            def __init__(self, name):
                self.name = name
        sch = Sch('sch')
        self.cfg.load_schedulers(self.filename,
                dict(schedulers=[sch]), self.errors)
        self.assertResults(schedulers=dict(sch=sch))


    def test_load_builders_defaults(self):
        self.cfg.load_builders(self.filename, {}, self.errors)
        self.assertResults(builders=[])

    def test_load_builders_not_list(self):
        self.cfg.load_builders(self.filename,
                dict(builders=dict()), self.errors)
        self.assertConfigError(self.errors, "must be a list")

    def test_load_builders_not_instance(self):
        self.cfg.load_builders(self.filename,
                dict(builders=[mock.Mock()]), self.errors)
        self.assertConfigError(self.errors, "must be a list of")

    def test_load_builders(self):
        bldr = config.BuilderConfig(name='x',
                        factory=mock.Mock(), slavename='x')
        self.cfg.load_builders(self.filename,
                dict(builders=[bldr]), self.errors)
        self.assertResults(builders=[bldr])

    def test_load_builders_dict(self):
        bldr = dict(name='x', factory=mock.Mock(), slavename='x')
        self.cfg.load_builders(self.filename,
                dict(builders=[bldr]), self.errors)
        self.assertIsInstance(self.cfg.builders[0], config.BuilderConfig)
        self.assertEqual(self.cfg.builders[0].name, 'x')


    def test_load_slaves_defaults(self):
        self.cfg.load_slaves(self.filename, {}, self.errors)
        self.assertResults(slaves=[])

    def test_load_slaves_not_list(self):
        self.cfg.load_slaves(self.filename,
                dict(slaves=dict()), self.errors)
        self.assertConfigError(self.errors, "must be a list")

    def test_load_slaves_not_instance(self):
        self.cfg.load_slaves(self.filename,
                dict(slaves=[mock.Mock()]), self.errors)
        self.assertConfigError(self.errors, "must be a list of")

    def test_load_slaves_reserved_names(self):
        for name in 'debug', 'change', 'status':
            self.cfg.load_slaves(self.filename,
                    dict(slaves=[buildslave.BuildSlave(name, 'x')]),
                    self.errors)
            self.assertConfigError(self.errors, "is reserved")
            self.errors.errors[:] = [] # clear out the errors

    def test_load_slaves(self):
        sl = buildslave.BuildSlave('foo', 'x')
        self.cfg.load_slaves(self.filename,
                dict(slaves=[sl]), self.errors)
        self.assertResults(slaves=[sl])


    def test_load_change_sources_defaults(self):
        self.cfg.load_change_sources(self.filename, {}, self.errors)
        self.assertResults(change_sources=[])

    def test_load_change_sources_not_instance(self):
        self.cfg.load_change_sources(self.filename,
                dict(change_source=[mock.Mock()]), self.errors)
        self.assertConfigError(self.errors, "must be a list of")

    def test_load_change_sources_single(self):
        chsrc = FakeChangeSource()
        self.cfg.load_change_sources(self.filename,
                dict(change_source=chsrc), self.errors)
        self.assertResults(change_sources=[chsrc])

    def test_load_change_sources_list(self):
        chsrc = FakeChangeSource()
        self.cfg.load_change_sources(self.filename,
                dict(change_source=[chsrc]), self.errors)
        self.assertResults(change_sources=[chsrc])


    def test_load_user_managers_defaults(self):
        self.cfg.load_user_managers(self.filename, {}, self.errors)
        self.assertResults(user_managers=[])

    def test_load_user_managers_not_list(self):
        self.cfg.load_user_managers(self.filename,
                dict(user_managers='foo'),
                self.errors)
        self.assertConfigError(self.errors, "must be a list")

    def test_load_user_managers(self):
        um = mock.Mock()
        self.cfg.load_user_managers(self.filename,
                dict(user_managers=[um]),
                self.errors)
        self.assertResults(user_managers=[um])

class MasterConfig_checkers(ConfigErrorsMixin, unittest.TestCase):

    def setUp(self):
        self.cfg = config.MasterConfig()
        self.errors = config.ConfigErrors()

    # utils

    def setup_basic_attrs(self):
        # set up a basic config for checking; this will be modified below
        sch = mock.Mock()
        sch.name = 'sch'
        sch.listBuilderNames = lambda : [ 'b1', 'b2' ]

        b1 = mock.Mock()
        b1.name = 'b1'

        b2 = mock.Mock()
        b2.name = 'b2'

        self.cfg.schedulers = dict(sch=sch)
        self.cfg.slaves = [ mock.Mock() ]
        self.cfg.builders = [ b1, b2 ]

    def setup_builder_locks(self, builder_lock=None, dup_builder_lock=False,
                                  step_lock=None, dup_step_lock=False):
        def bldr(name):
            b = mock.Mock()
            b.name = name
            b.locks = []
            b.factory.steps = [ ('cls', dict(locks=[])) ]
            return b

        def lock(name):
            l = mock.Mock()
            l.name = name
            return l

        b1, b2 = bldr('b1'), bldr('b2')
        self.cfg.builders = [ b1, b2 ]
        if builder_lock:
            b1.locks.append(lock(builder_lock))
            if dup_builder_lock:
                b2.locks.append(lock(builder_lock))
        if step_lock:
            s1, s2 = b1.factory.steps[0][1], b2.factory.steps[0][1]
            s1['locks'].append(lock(step_lock))
            if dup_step_lock:
                s2['locks'].append(lock(step_lock))

    # tests

    def test_check_single_master_multimaster(self):
        self.cfg.multiMaster = True
        self.cfg.check_single_master(self.errors)
        self.assertNoConfigErrors(self.errors)

    def test_check_single_master_no_builders(self):
        self.setup_basic_attrs()
        self.cfg.builders = [ ]
        self.cfg.check_single_master(self.errors)
        self.assertConfigError(self.errors, "no builders are configured")

    def test_check_single_master_no_slaves(self):
        self.setup_basic_attrs()
        self.cfg.slaves = [ ]
        self.cfg.check_single_master(self.errors)
        self.assertConfigError(self.errors, "no slaves are configured")

    def test_check_single_master_unsch_builder(self):
        self.setup_basic_attrs()
        b3 = mock.Mock()
        b3.name = 'b3'
        self.cfg.builders.append(b3)
        self.cfg.check_single_master(self.errors)
        self.assertConfigError(self.errors, "have no schedulers to drive them")


    def test_check_schedulers_unknown_builder(self):
        self.setup_basic_attrs()
        del self.cfg.builders[1] # remove b2, leaving b1

        self.cfg.check_schedulers(self.errors)
        self.assertConfigError(self.errors, "Unknown builder 'b2'")

    def test_check_schedulers(self):
        self.setup_basic_attrs()
        self.cfg.check_schedulers(self.errors)
        self.assertNoConfigErrors(self.errors)


    def test_check_locks_step_and_builder(self):
        self.setup_builder_locks(builder_lock='l', step_lock='l')
        self.cfg.check_locks(self.errors)
        self.assertConfigError(self.errors, "Two locks share")

    def test_check_locks_dup_builder_lock(self):
        self.setup_builder_locks(builder_lock='l', dup_builder_lock=True)
        self.cfg.check_locks(self.errors)
        self.assertConfigError(self.errors, "Two locks share")

    def test_check_locks_dup_step_lock(self):
        self.setup_builder_locks(step_lock='l', dup_step_lock=True)
        self.cfg.check_locks(self.errors)
        self.assertConfigError(self.errors, "Two locks share")

    def test_check_locks(self):
        self.setup_builder_locks(builder_lock='bl', step_lock='sl')
        self.cfg.check_locks(self.errors)
        self.assertNoConfigErrors(self.errors)

    def test_check_locks_none(self):
        # no locks in the whole config, should be fine
        self.setup_builder_locks()
        self.cfg.check_locks(self.errors)
        self.assertNoConfigErrors(self.errors)


    def test_check_builders_unknown_slave(self):
        sl = mock.Mock()
        sl.slavename = 'xyz'
        self.cfg.slaves = [ sl ]

        b1 = FakeBuilder(slavenames=[ 'xyz', 'abc' ], builddir='x', name='b1')
        self.cfg.builders = [ b1 ]

        self.cfg.check_builders(self.errors)
        self.assertConfigError(self.errors,
                "builder 'b1' uses unknown slaves 'abc'")

    def test_check_builders_duplicate_name(self):
        b1 = FakeBuilder(slavenames=[], name='b1', builddir='1')
        b2 = FakeBuilder(slavenames=[], name='b1', builddir='2')
        self.cfg.builders = [ b1, b2 ]

        self.cfg.check_builders(self.errors)
        self.assertConfigError(self.errors,
                "duplicate builder name 'b1'")

    def test_check_builders_duplicate_builddir(self):
        b1 = FakeBuilder(slavenames=[], name='b1', builddir='dir')
        b2 = FakeBuilder(slavenames=[], name='b2', builddir='dir')
        self.cfg.builders = [ b1, b2 ]

        self.cfg.check_builders(self.errors)
        self.assertConfigError(self.errors,
                "duplicate builder builddir 'dir'")

    def test_check_builders(self):
        sl = mock.Mock()
        sl.slavename = 'a'
        self.cfg.slaves = [ sl ]

        b1 = FakeBuilder(slavenames=[ 'a' ], name='b1', builddir='dir1')
        b2 = FakeBuilder(slavenames=[ 'a' ], name='b2', builddir='dir2')
        self.cfg.builders = [ b1, b2 ]

        self.cfg.check_builders(self.errors)
        self.assertNoConfigErrors(self.errors)


    def test_check_status_fails(self):
        st = FakeStatusReceiver()
        st.checkConfig = lambda status, errors : errors.addError("oh noes")
        self.cfg.status = [ st ]

        self.cfg.check_status(self.errors)

        self.assertConfigError(self.errors, "oh noes")

    def test_check_status(self):
        st = FakeStatusReceiver()
        st.checkConfig = mock.Mock()
        self.cfg.status = [ st ]

        self.cfg.check_status(self.errors)

        self.assertNoConfigErrors(self.errors)
        st.checkConfig.assert_called_once_with(self.cfg.status, self.errors)

    def test_check_horizons(self):
        self.cfg.logHorizon = 100
        self.cfg.buildHorizon = 50
        self.cfg.check_horizons(self.errors)

        self.assertConfigError(self.errors, "logHorizon must be less")

    def test_check_slavePortnum_set(self):
        self.cfg.slavePortnum = 10
        self.cfg.check_slavePortnum(self.errors)
        self.assertNoConfigErrors(self.errors)

    def test_check_slavePortnum_not_set_slaves(self):
        self.cfg.slaves = [ mock.Mock() ]
        self.cfg.check_slavePortnum(self.errors)
        self.assertConfigError(self.errors,
                "slaves are configured, but no slavePortnum is set")

    def test_check_slavePortnum_not_set_debug(self):
        self.cfg.debugPassword = 'ssh'
        self.cfg.check_slavePortnum(self.errors)
        self.assertConfigError(self.errors,
                "debug client is configured, but no slavePortnum is set")


class BuilderConfig(ConfigErrorsMixin, unittest.TestCase):

    factory = mock.Mock()

    # utils

    def assertAttributes(self, cfg, **expected):
        got = dict([
            (attr, getattr(cfg, attr))
            for attr, exp in expected.iteritems() ])
        self.assertEqual(got, expected)

    # tests

    def test_no_name(self):
        self.assertRaisesConfigError(
            "builder's name is required",
            lambda : config.BuilderConfig(
                factory=self.factory, slavenames=['a']))

    def test_reserved_name(self):
        self.assertRaisesConfigError(
            "builder names must not start with an underscore: '_a'",
            lambda : config.BuilderConfig(name='_a',
                factory=self.factory, slavenames=['a']))

    def test_no_factory(self):
        self.assertRaisesConfigError(
            "builder 'a' has no factory",
            lambda : config.BuilderConfig(
                name='a', slavenames=['a']))

    def test_no_slavenames(self):
        self.assertRaisesConfigError(
            "builder 'a': at least one slavename is required",
            lambda : config.BuilderConfig(
                name='a', factory=self.factory))

    def test_bogus_slavenames(self):
        self.assertRaisesConfigError(
            "slavenames must be a list or a string",
            lambda : config.BuilderConfig(
                name='a', slavenames={1:2}, factory=self.factory))

    def test_bogus_slavename(self):
        self.assertRaisesConfigError(
            "slavename must be a string",
            lambda : config.BuilderConfig(
                name='a', slavename=1, factory=self.factory))

    def test_bogus_category(self):
        self.assertRaisesConfigError(
            "category must be a string",
            lambda : config.BuilderConfig(category=13,
                name='a', slavenames=['a'], factory=self.factory))

    def test_defaults(self):
        cfg = config.BuilderConfig(
            name='a b c', slavename='a', factory=self.factory)
        self.assertIdentical(cfg.factory, self.factory)
        self.assertAttributes(cfg,
            name='a b c',
            slavenames=['a'],
            builddir='a_b_c',
            slavebuilddir='a_b_c',
            category='',
            nextSlave=None,
            locks=[],
            env={},
            properties={},
            mergeRequests=None)

    def test_args(self):
        cfg = config.BuilderConfig(
            name='b', slavename='s1', slavenames='s2', builddir='bd',
            slavebuilddir='sbd', factory=self.factory, category='c',
            nextSlave=lambda : 'ns', nextBuild=lambda : 'nb', locks=['l'],
            env=dict(x=10), properties=dict(y=20), mergeRequests='mr')
        self.assertIdentical(cfg.factory, self.factory)
        self.assertAttributes(cfg,
            name='b',
            slavenames=['s2', 's1'],
            builddir='bd',
            slavebuilddir='sbd',
            category='c',
            locks=['l'],
            env={'x':10},
            properties={'y':20},
            mergeRequests='mr')

    def test_getConfigDict(self):
        ns = lambda : 'ns'
        nb = lambda : 'nb'
        cfg = config.BuilderConfig(
            name='b', slavename='s1', slavenames='s2', builddir='bd',
            slavebuilddir='sbd', factory=self.factory, category='c',
            nextSlave=ns, nextBuild=nb, locks=['l'],
            env=dict(x=10), properties=dict(y=20), mergeRequests='mr')
        self.assertEqual(cfg.getConfigDict(), {'builddir': 'bd',
            'category': 'c',
            'env': {'x': 10},
            'factory': self.factory,
            'locks': ['l'],
            'mergeRequests': 'mr',
            'name': 'b',
            'nextBuild': nb,
            'nextSlave': ns,
            'properties': {'y': 20},
            'slavebuilddir': 'sbd',
            'slavenames': ['s2', 's1'],
        })



class FakeService(config.ReconfigurableServiceMixin,
                    service.Service):

    succeed = True
    call_index = 1

    def reconfigService(self, new_config):
        self.called = FakeService.call_index
        FakeService.call_index += 1
        d = config.ReconfigurableServiceMixin.reconfigService(self, new_config)
        if not self.succeed:
            @d.addCallback
            def fail(_):
                raise ValueError("oh noes")
        return d



class FakeMultiService(config.ReconfigurableServiceMixin,
                    service.MultiService):

    def reconfigService(self, new_config):
        self.called = True
        d = config.ReconfigurableServiceMixin.reconfigService(self, new_config)
        return d



class ReconfigurableServiceMixin(unittest.TestCase):

    def test_service(self):
        svc = FakeService()
        d = svc.reconfigService(mock.Mock())
        @d.addCallback
        def check(_):
            self.assertTrue(svc.called)
        return d

    @defer.deferredGenerator
    def test_service_failure(self):
        svc = FakeService()
        svc.succeed = False
        try:
            wfd = defer.waitForDeferred(
                svc.reconfigService(mock.Mock()))
            yield wfd
            wfd.getResult()
        except ValueError:
            pass
        else:
            self.fail("should have raised ValueError")

    def test_multiservice(self):
        svc = FakeMultiService()
        ch1 = FakeService()
        ch1.setServiceParent(svc)
        ch2 = FakeMultiService()
        ch2.setServiceParent(svc)
        ch3 = FakeService()
        ch3.setServiceParent(ch2)
        d = svc.reconfigService(mock.Mock())
        @d.addCallback
        def check(_):
            self.assertTrue(svc.called)
            self.assertTrue(ch1.called)
            self.assertTrue(ch2.called)
            self.assertTrue(ch3.called)
        return d

    def test_multiservice_priority(self):
        parent = FakeMultiService()
        svc128 = FakeService()
        svc128.setServiceParent(parent)

        services = [ svc128 ]
        for i in range(20, 1, -1):
            svc = FakeService()
            svc.reconfig_priority = i
            svc.setServiceParent(parent)
            services.append(svc)

        d = parent.reconfigService(mock.Mock())
        @d.addCallback
        def check(_):
            prio_order = [ svc.called for svc in services ]
            called_order = sorted(prio_order)
            self.assertEqual(prio_order, called_order)
        return d

    @compat.usesFlushLoggedErrors
    @defer.deferredGenerator
    def test_multiservice_nested_failure(self):
        svc = FakeMultiService()
        ch1 = FakeService()
        ch1.setServiceParent(svc)
        ch1.succeed = False
        try:
            wfd = defer.waitForDeferred(
                svc.reconfigService(mock.Mock()))
            yield wfd
            wfd.getResult()
        except ValueError:
            pass
        else:
            self.fail("should have raised ValueError")



import pkg_resources
import re

from genshi.builder import tag

from trac.core import Component, implements
from trac.config import Option
from trac.wiki.formatter import format_to_oneliner
from trac.util.datefmt import to_timestamp, to_datetime
from trac.mimeview.api import Context

# Interfaces
from trac.timeline.api import ITimelineEventProvider
from trac.web import IRequestHandler
from trac.web.chrome import INavigationContributor, ITemplateProvider

from api import BuildBotSystem

class TracBuildBotWatcher(Component):
	implements(ITimelineEventProvider, IRequestHandler, ITemplateProvider,
		INavigationContributor)
	buildbot_url = Option('bbwatcher', 'buildmaster', '127.0.0.1:8010',
		'The location of the BuildBot webserver. Do not include the /xmlrpc')

	BUILDER_REGEX = r'/buildbot/builder(?:/(.+))?$'
	BUILDER_RE = re.compile(BUILDER_REGEX)
	# Template Provider
	def get_htdocs_dirs(self):
		return []
	def get_templates_dirs(self):
		return [pkg_resources.resource_filename('bbwatcher', 'templates')]

	# Nav Contributor
	def get_active_navigation_item(self, req):
		return 'buildbot'
	def get_navigation_items(self, req):
		yield 'mainnav', 'buildbot', tag.a('BuildBot',href=req.href.buildbot())

	# Timeline Methods
	def get_timeline_filters(self, req):
		yield  ('bbwatcher', 'Builds', False)

	def get_timeline_events(self, req, start, stop, filters):
		#if not 'bbwatcher' in filters:
		#	return
		try:
			master = BuildBotSystem(self.buildbot_url)
		except Exception, e:
			print 'Error hitting BuildBot', e
			return
		# This was a comprehension: the loop is clearer
		for build in master.getAllBuildsInInterval(to_timestamp(start), to_timestamp(stop)):
			# BuildBot builds are reported as
			# (builder_name, num, end, branch, rev, results, text)
			print 'Reporting build', build
			yield ('build', to_datetime(build[2]), '', build)

	def render_timeline_event(self, context, field, event):
		builder_name, num, end, branch, rev, results, text = event[3]
		if field == 'url':
			return None
		elif field == 'title':
			return tag('Build ', tag.a('#%s'%num, href=context.href.buildbot('builder/%s/%s'%(builder_name, num))),
				' of ', builder_name, ' ', results == 'success' and tag.span('passed', style="color: #080") or tag.span('failed', style="color: #f00"))
		elif field == 'description':
			return format_to_oneliner(self.env, context, 'Built from %s'%(rev and 'r%s sources'%rev or 'local changes (see TryBuildUsage)'))

	# RequestHandler
	def _handle_builder(self, req):
		m = self.BUILDER_RE.match(req.path_info)
		try:
			builder = m.group(1) or None
		except Exception, e:
			builder = None
		master = BuildBotSystem(self.buildbot_url)
		if builder is None:
			data = { 'names': master.getAllBuilders() }
			return 'bbw_allbuilders.html', data, 'text/html'
		else:
			class Foo:
				pass
			b = Foo()
			b.name = str(builder)
			b.current = 'CURRENT-TEXT'
			b.recent = []
			b.slaves = []
			data = { 'builder': b }
			try:
				master = BuildBotSystem(self.buildbot_url)
				data = { 'builder': master.getBuilder(builder) }
			except Exception, e:
				print 'Error fetching builder stats', e
			data['context'] = Context.from_request(req, ('buildbot', builder))
			return 'bbw_builder.html', data, 'text/html'

	def match_request(self, req):
		return req.path_info.startswith('/buildbot') and 1 or 0

	def process_request(self, req):
		if req.path_info.startswith('/buildbot/builder'):
			return self._handle_builder(req)
		return 'bbw_welcome.html', { 'url': self.buildbot_url }, 'text/html'
